from __future__ import annotations

import operator
from collections.abc import AsyncIterator, Callable
from functools import cached_property
from typing import (
    Any,
    Iterator,
    Literal,
    Optional,
    Self,
    Sequence,
    TypeVar,
    overload,
)

from pydantic import TypeAdapter
from sqlalchemy import (
    FrozenResult,
    IteratorResult,
    Result,
    Row,
    ScalarResult,
)
from sqlalchemy.engine.result import (
    _NO_ROW,
    _R,
    FilterResult,
    ResultMetaData,
    _KeyIndexType,
    _UniqueFilterType,
    _WithKeys,
)
from sqlalchemy.engine.row import RowMapping
from sqlalchemy.sql.base import _generative
from sqlalchemy.util.concurrency import greenlet_spawn

from arcanus.utils import get_cached_adapter

_T = TypeVar("_T", bound=Any)
_TP = TypeVar("_TP", bound=tuple[Any, ...])


class AdaptedCommon(FilterResult[_R]):
    __slots__ = ()

    _real_result: Result[Any]
    _metadata: ResultMetaData

    def close(self) -> None:
        self._real_result.close()

    @property
    def closed(self) -> bool:
        return self._real_result.closed


class AdaptedResult(_WithKeys, AdaptedCommon[Row[_TP]]):
    entities: tuple[type[Any], ...]

    _real_result: Result[_TP]
    _row_logging_fn: Optional[Callable[[Row[Any]], Row[Any]]] = None

    def __init__(
        self,
        real_result: Result[_TP],
        entities: tuple[type[Any], ...] = (),
    ):
        self._real_result = real_result

        self._metadata = real_result._metadata
        self._unique_filter_state = real_result._unique_filter_state
        self._source_supports_scalars = real_result._source_supports_scalars
        self._post_creational_filter = None

        self.entities = entities

        # BaseCursorResult pre-generates the "_row_getter".  Use that
        # if available rather than building a second one
        if "_row_getter" in real_result.__dict__:
            self._set_memoized_attribute(
                "_row_getter", real_result.__dict__["_row_getter"]
            )

    @cached_property
    def adapter(self) -> TypeAdapter:
        """Get the adapter for tuple results."""
        return get_cached_adapter(tuple[*self.entities])

    @cached_property
    def scalar_adapter(self) -> TypeAdapter:
        """Get the adapter for scalar results."""
        return (
            get_cached_adapter(self.entities[0])
            if self.entities
            else TypeAdapter(object)
        )

    def _adapt(self, row: Any) -> Any:
        """Adapt a row using TypeAdapter validation."""
        return self.adapter.validate_python(row)

    def _adapt_scalar(self, item: Any) -> Any:
        """Adapt a scalar result using TypeAdapter validation."""
        return self.scalar_adapter.validate_python(item)

    @property
    def t(self) -> Self:
        return self

    @property
    def tuples(self) -> Self:
        return self

    @_generative
    def unique(self, strategy: Optional[_UniqueFilterType] = None) -> Self:
        """Apply unique filtering to the objects returned by this
        :class:`_asyncio.AsyncResult`.

        Refer to :meth:`_engine.Result.unique` in the synchronous
        SQLAlchemy API for a complete behavioral description.

        """
        self._unique_filter_state = (set(), strategy)
        return self

    def columns(self, *col_expressions: _KeyIndexType) -> Self:
        r"""Establish the columns that should be returned in each row.

        Refer to :meth:`_engine.Result.columns` in the synchronous
        SQLAlchemy API for a complete behavioral description.

        """
        return self._column_slices(col_expressions)

    def __iter__(self) -> Iterator[Row[_TP]]:
        for row in self._iter_impl():
            yield self._adapt(row)

    def __next__(self) -> Row[_TP]:
        return self._adapt(self._next_impl())

    def partitions(self, size: int | None = None) -> Iterator[Sequence[Row[_TP]]]:
        while True:
            partition = self._manyrow_getter(self, size)
            if partition:
                yield [self._adapt(row) for row in partition]
            else:
                break

    def fetchall(self) -> Sequence[Row[_TP]]:
        return [self._adapt(row) for row in self._allrows()]

    def fetchone(self) -> Optional[Row[_TP]]:
        row = self._onerow_getter(self)
        if row is _NO_ROW:
            return None
        else:
            return self._adapt(row)

    def fetchmany(self, size: int | None = None) -> Sequence[Row[_TP]]:
        return [self._adapt(row) for row in self._manyrow_getter(size)]

    def all(self) -> Sequence[Row[_TP]]:
        return [self._adapt(row) for row in self._allrows()]

    def first(self) -> Optional[Row[_TP]]:
        return (
            self._adapt(row)
            if (
                row := self._only_one_row(
                    raise_for_second_row=False,
                    raise_for_none=False,
                    scalar=False,
                )
            )
            else None
        )

    def one(self) -> Row[_TP]:
        return self._adapt(
            self._only_one_row(
                raise_for_second_row=True,
                raise_for_none=True,
                scalar=False,
            )
        )

    def one_or_none(self) -> Optional[Row[_TP]]:
        return (
            self._adapt(row)
            if (
                row := self._only_one_row(
                    raise_for_second_row=True,
                    raise_for_none=False,
                    scalar=False,
                )
            )
            else None
        )

    @overload
    def scalar(self: AdaptedResult[tuple[_T]]) -> Optional[_T]: ...

    @overload
    def scalar(self) -> Any: ...
    def scalar(self) -> Any:
        return self._adapt_scalar(
            self._only_one_row(
                raise_for_second_row=False,
                raise_for_none=False,
                scalar=True,
            )
        )

    @overload
    def scalar_one(self: AdaptedResult[tuple[_T]]) -> _T: ...

    @overload
    def scalar_one(self) -> Any: ...
    def scalar_one(self) -> Any:
        return self._adapt_scalar(
            self._only_one_row(
                raise_for_second_row=True,
                raise_for_none=True,
                scalar=True,
            )
        )

    @overload
    def scalar_one_or_none(self: AdaptedResult[tuple[_T]]) -> Optional[_T]: ...

    @overload
    def scalar_one_or_none(self) -> Optional[Any]: ...

    def scalar_one_or_none(self) -> Optional[Any]:
        return (
            self._adapt_scalar(row)
            if (
                row := self._only_one_row(
                    raise_for_second_row=True,
                    raise_for_none=False,
                    scalar=True,
                )
            )
            else None
        )

    @overload
    def scalars(self: AdaptedResult[tuple[_T]]) -> AdaptedScalarResult[_T]: ...

    @overload
    def scalars(
        self: Result[tuple[_T]], index: Literal[0]
    ) -> AdaptedScalarResult[_T]: ...

    @overload
    def scalars(self, index: _KeyIndexType = 0) -> AdaptedScalarResult[Any]: ...
    def scalars(self, index: _KeyIndexType = 0) -> AdaptedScalarResult[Any]:
        return AdaptedScalarResult(
            self,
            index=index,
            entities=self.entities,
        )

    def freeze(self) -> AdaptedFrozenResult[_TP]:
        return AdaptedFrozenResult(
            self,
            entities=self.entities,
        )

    def mappings(self) -> AdaptedMappingResult:
        return AdaptedMappingResult(self)


class AdaptedScalarResult(ScalarResult[_R]):
    _entities: tuple[type[Any], ...]

    __slots__ = ()

    _generate_rows = False
    _real_result: AdaptedResult[Any]

    def __init__(
        self,
        real_result: AdaptedResult[Any],
        index: _KeyIndexType,
        entities: tuple[type[Any], ...] = (),
    ):
        self._unique_filter_state = real_result._unique_filter_state
        self._real_result = real_result

        if real_result._source_supports_scalars:
            self._metadata = real_result._metadata
            self._post_creational_filter = None
        else:
            self._metadata = real_result._metadata._reduce([index])
            self._post_creational_filter = operator.itemgetter(0)

        self._entities = entities

    @cached_property
    def scalar_adapter(self) -> TypeAdapter:
        """Get the adapter for scalar results."""
        return (
            get_cached_adapter(self._entities[0])
            if self._entities
            else TypeAdapter(object)
        )

    def _adapt_scalar(self, item: Any) -> Any:
        """Adapt a scalar result using TypeAdapter validation."""
        return self.scalar_adapter.validate_python(item)

    def unique(
        self,
        strategy: Optional[_UniqueFilterType] = None,
    ) -> Self:
        self._unique_filter_state = (set(), strategy)
        return self

    def __iter__(self) -> Iterator[_R]:
        for row in self._iter_impl():
            yield self._adapt_scalar(row)

    def __next__(self) -> _R:
        return self._adapt_scalar(self._next_impl())

    def partitions(self, size: int | None = None) -> Iterator[Sequence[_R]]:
        while True:
            partition = self._manyrow_getter(self, size)
            if partition:
                yield [self._adapt_scalar(scalar) for scalar in partition]
            else:
                break

    def fetchall(self) -> Sequence[_R]:
        return [self._adapt_scalar(scalar) for scalar in self._allrows()]

    def fetchmany(self, size: int | None = None) -> Sequence[_R]:
        return [self._adapt_scalar(scalar) for scalar in self._manyrow_getter(size)]

    def all(self) -> Sequence[_R]:
        return [self._adapt_scalar(scalar) for scalar in self._allrows()]

    def first(self) -> _R | None:
        return (
            self._adapt_scalar(row)
            if (
                row := self._only_one_row(
                    raise_for_second_row=False,
                    raise_for_none=False,
                    scalar=False,
                )
            )
            else None
        )

    def one(self) -> _R:
        return self._adapt_scalar(
            self._only_one_row(
                raise_for_second_row=True,
                raise_for_none=True,
                scalar=False,
            )
        )

    def one_or_none(self) -> _R | None:
        return (
            self._adapt_scalar(row)
            if (
                row := self._only_one_row(
                    raise_for_second_row=True,
                    raise_for_none=False,
                    scalar=False,
                )
            )
            else None
        )


class AdaptedFrozenResult(FrozenResult[_TP]):
    data: Sequence[Any]
    _entities: tuple[type[Any], ...]

    def __init__(
        self,
        result: AdaptedResult[_TP],
        entities: tuple[type[Any], ...] = (),
    ):
        self.metadata = result._metadata._for_freeze()
        self._source_supports_scalars = result._source_supports_scalars
        self._attributes = result._attributes
        self._entities = entities

        if self._source_supports_scalars:
            self.data = self._adapt_scalar(result._real_result._raw_row_iterator())
        else:
            self.data = result.fetchall()

    @cached_property
    def adapter(self) -> TypeAdapter:
        """Get the adapter for tuple results."""
        return get_cached_adapter(tuple[*self._entities])

    @cached_property
    def scalar_adapter(self) -> TypeAdapter:
        """Get the adapter for scalar results."""
        return (
            get_cached_adapter(self._entities[0])
            if self._entities
            else TypeAdapter(object)
        )

    def _adapt_scalar(self, item: Any) -> Any:
        """Adapt a scalar result using TypeAdapter validation."""
        return self.scalar_adapter.validate_python(item)

    def rewrite_rows(self) -> Sequence[Sequence[Any]]:
        if self._source_supports_scalars:
            return [[elem] for elem in self.data]
        else:
            return [list(row) for row in self.data]

    def with_new_rows(self, tuple_data: Sequence[Row[_TP]]) -> AdaptedFrozenResult[_TP]:
        afr = AdaptedFrozenResult.__new__(AdaptedFrozenResult)
        afr.metadata = self.metadata
        afr._attributes = self._attributes
        afr._source_supports_scalars = self._source_supports_scalars
        afr._entities = self._entities

        if self._source_supports_scalars:
            afr.data = [d[0] for d in tuple_data]
        else:
            afr.data = tuple_data
        return afr

    def __call__(self) -> Result[_TP]:
        result = IteratorResult(
            self.metadata,
            iter(self.data),
        )
        result._attributes = self._attributes  # type: ignore
        result._source_supports_scalars = self._source_supports_scalars
        return result


class AdaptedMappingResult(_WithKeys, AdaptedCommon[RowMapping]):
    __slots__ = ()

    _generate_rows = True
    _post_creational_filter = operator.attrgetter("_mapping")

    _real_result: AdaptedResult[Any]
    _entities: tuple[type[Any], ...]

    def __init__(self, result: AdaptedResult[Any]):
        self._real_result = result
        self._metadata = result._metadata
        self._unique_filter_state = result._unique_filter_state
        if result._source_supports_scalars:
            self._metadata = self._metadata._reduce([0])

        self._entities = result.entities

    @cached_property
    def adapter(self) -> TypeAdapter:
        """Get the adapter for tuple results."""
        return get_cached_adapter(tuple[*self._entities])

    @cached_property
    def scalar_adapter(self) -> TypeAdapter:
        """Get the adapter for scalar results."""
        return (
            get_cached_adapter(self._entities[0])
            if self._entities
            else TypeAdapter(object)
        )

    def _adapt(self, row: Any) -> Any:
        """Adapt a row using TypeAdapter validation."""
        return self.adapter.validate_python(row)

    @_generative
    def unique(self, strategy: Optional[_UniqueFilterType] = None) -> Self:
        self._unique_filter_state = (set(), strategy)
        return self

    def columns(self, *col_expressions: _KeyIndexType) -> Self:
        return self._column_slices(col_expressions)

    def partitions(self, size: int | None = None) -> Iterator[Sequence[dict[str, Any]]]:
        while True:
            partition = self._manyrow_getter(self, size)
            if partition:
                yield list(
                    dict(zip(row.keys(), self._adapt(row.values())))
                    for row in partition
                )
            else:
                break

    def fetchone(self) -> Optional[dict[str, Any]]:
        row = self._onerow_getter(self)
        if row is _NO_ROW:
            return None
        else:
            return dict(zip(row.keys(), self._adapt(row.values())))

    def fetchmany(self, size: int | None = None) -> Sequence[dict[str, Any]]:
        return list(
            dict(zip(row.keys(), self._adapt(row.values())))
            for row in self._manyrow_getter(size)
        )

    def fetchall(self) -> Sequence[dict[str, Any]]:
        """A synonym for the :meth:`AdaptedMappingResult.all` method."""
        return list(
            dict(zip(row.keys(), self._adapt(row.values()))) for row in self._allrows()
        )

    def first(self) -> dict[str, Any] | None:
        return (
            dict(zip(row.keys(), self._adapt(row.values())))
            if (
                row := self._only_one_row(
                    raise_for_second_row=False,
                    raise_for_none=False,
                    scalar=False,
                )
            )
            else None
        )

    def all(self) -> Sequence[dict[str, Any]]:
        return list(
            dict(zip(row.keys(), self._adapt(row.values()))) for row in self._allrows()
        )

    def one(self) -> dict[str, Any]:
        row = self._only_one_row(
            raise_for_second_row=True,
            raise_for_none=True,
            scalar=False,
        )
        return dict(zip(row.keys(), self._adapt(row.values())))

    def one_or_none(self) -> dict[str, Any] | None:
        return (
            dict(zip(row.keys(), self._adapt(row.values())))
            if (
                row := self._only_one_row(
                    raise_for_second_row=True,
                    raise_for_none=False,
                    scalar=False,
                )
            )
            else None
        )


class AsyncAdaptedCommon(FilterResult[_R]):
    """Base class for async adapted results with common functionality."""

    __slots__ = ()

    _real_result: Result[Any]
    _metadata: ResultMetaData

    async def close(self) -> None:
        """Close this result."""
        await greenlet_spawn(self._real_result.close)

    @property
    def closed(self) -> bool:
        return self._real_result.closed


class AsyncAdaptedResult(_WithKeys, AsyncAdaptedCommon[Row[_TP]]):
    """An asyncio wrapper around an AdaptedResult that provides async iteration
    and adapted row conversion for transmuter types.
    """

    __slots__ = ()

    entities: tuple[type[Any], ...]
    _real_result: Result[_TP]

    def __init__(
        self,
        real_result: Result[_TP],
        entities: tuple[type[Any], ...] = (),
    ):
        self._real_result = real_result

        self._metadata = real_result._metadata
        self._unique_filter_state = real_result._unique_filter_state
        self._source_supports_scalars = real_result._source_supports_scalars
        self._post_creational_filter = None

        self.entities = entities

        # BaseCursorResult pre-generates the "_row_getter".  Use that
        # if available rather than building a second one
        if "_row_getter" in real_result.__dict__:
            self._set_memoized_attribute(
                "_row_getter", real_result.__dict__["_row_getter"]
            )

    @cached_property
    def adapter(self) -> TypeAdapter:
        """Get the adapter for tuple results."""
        return get_cached_adapter(tuple[*self.entities])

    @cached_property
    def scalar_adapter(self) -> TypeAdapter:
        """Get the adapter for scalar results."""
        return (
            get_cached_adapter(self.entities[0])
            if self.entities
            else TypeAdapter(object)
        )

    def _adapt(self, row: Any) -> Any:
        """Adapt a row using TypeAdapter validation."""
        return self.adapter.validate_python(row)

    def _adapt_scalar(self, item: Any) -> Any:
        """Adapt a scalar result using TypeAdapter validation."""
        return self.scalar_adapter.validate_python(item)

    @property
    def t(self) -> Self:
        return self

    def tuples(self) -> Self:
        return self

    @_generative
    def unique(self, strategy: Optional[_UniqueFilterType] = None) -> Self:
        """Apply unique filtering to the objects returned by this
        :class:`AsyncAdaptedResult`.
        """
        self._unique_filter_state = (set(), strategy)
        return self

    def columns(self, *col_expressions: _KeyIndexType) -> Self:
        r"""Establish the columns that should be returned in each row."""
        return self._column_slices(col_expressions)

    async def partitions(
        self, size: int | None = None
    ) -> AsyncIterator[Sequence[Row[_TP]]]:
        """Iterate through sub-lists of rows of the size given."""
        getter = self._manyrow_getter

        while True:
            partition = await greenlet_spawn(getter, self, size)
            if partition:
                yield [self._adapt(row) for row in partition]
            else:
                break

    async def fetchall(self) -> Sequence[Row[_TP]]:
        """A synonym for the :meth:`AsyncAdaptedResult.all` method."""
        rows = await greenlet_spawn(self._allrows)
        return [self._adapt(row) for row in rows]

    async def fetchone(self) -> Optional[Row[_TP]]:
        """Fetch one row."""
        row = await greenlet_spawn(self._onerow_getter, self)
        if row is _NO_ROW:
            return None
        else:
            return self._adapt(row)

    async def fetchmany(self, size: int | None = None) -> Sequence[Row[_TP]]:
        """Fetch many rows."""
        rows = await greenlet_spawn(self._manyrow_getter, self, size)
        return [self._adapt(row) for row in rows]

    async def all(self) -> Sequence[Row[_TP]]:
        """Return all rows in a list."""
        rows = await greenlet_spawn(self._allrows)
        return [self._adapt(row) for row in rows]

    def __aiter__(self) -> AsyncAdaptedResult[_TP]:
        return self

    async def __anext__(self) -> Row[_TP]:
        row = await greenlet_spawn(self._onerow_getter, self)
        if row is _NO_ROW:
            raise StopAsyncIteration()
        else:
            return self._adapt(row)

    async def first(self) -> Optional[Row[_TP]]:
        """Fetch the first row or ``None`` if no row is present."""
        row = await greenlet_spawn(self._only_one_row, False, False, False)
        return self._adapt(row) if row else None

    async def one_or_none(self) -> Optional[Row[_TP]]:
        """Return at most one result or raise an exception."""
        row = await greenlet_spawn(self._only_one_row, True, False, False)
        return self._adapt(row) if row else None

    async def one(self) -> Row[_TP]:
        """Return exactly one row or raise an exception."""
        row = await greenlet_spawn(self._only_one_row, True, True, False)
        return self._adapt(row)

    @overload
    async def scalar_one(self: AsyncAdaptedResult[tuple[_T]]) -> _T: ...

    @overload
    async def scalar_one(self) -> Any: ...

    async def scalar_one(self) -> Any:
        """Return exactly one scalar result or raise an exception."""
        scalar = await greenlet_spawn(self._only_one_row, True, True, True)
        return self._adapt_scalar(scalar)

    @overload
    async def scalar_one_or_none(
        self: AsyncAdaptedResult[tuple[_T]],
    ) -> Optional[_T]: ...

    @overload
    async def scalar_one_or_none(self) -> Optional[Any]: ...

    async def scalar_one_or_none(self) -> Optional[Any]:
        """Return exactly one scalar result or ``None``."""
        scalar = await greenlet_spawn(self._only_one_row, True, False, True)
        return self._adapt_scalar(scalar) if scalar else None

    @overload
    async def scalar(self: AsyncAdaptedResult[tuple[_T]]) -> Optional[_T]: ...

    @overload
    async def scalar(self) -> Any: ...

    async def scalar(self) -> Any:
        """Fetch the first column of the first row, and close the result set."""
        scalar = await greenlet_spawn(self._only_one_row, False, False, True)
        return self._adapt_scalar(scalar) if scalar else None

    async def freeze(self) -> AdaptedFrozenResult[_TP]:
        """Return a callable object that will produce copies of this result."""
        return await greenlet_spawn(
            AdaptedFrozenResult, self._real_result, self.entities
        )

    @overload
    def scalars(
        self: AsyncAdaptedResult[tuple[_T]], index: Literal[0]
    ) -> AsyncAdaptedScalarResult[_T]: ...

    @overload
    def scalars(
        self: AsyncAdaptedResult[tuple[_T]],
    ) -> AsyncAdaptedScalarResult[_T]: ...

    @overload
    def scalars(self, index: _KeyIndexType = 0) -> AsyncAdaptedScalarResult[Any]: ...

    def scalars(self, index: _KeyIndexType = 0) -> AsyncAdaptedScalarResult[Any]:
        """Return an :class:`AsyncAdaptedScalarResult` filtering object."""
        return AsyncAdaptedScalarResult(self._real_result, index, self.entities)

    def mappings(self) -> AsyncAdaptedMappingResult:
        """Apply a mappings filter to returned rows."""
        return AsyncAdaptedMappingResult(self._real_result, self.entities)


class AsyncAdaptedScalarResult(AsyncAdaptedCommon[_R]):
    """An async wrapper that returns adapted scalar values rather than Row values."""

    __slots__ = ()

    _generate_rows = False
    _entities: tuple[type[Any], ...]

    def __init__(
        self,
        real_result: Result[Any],
        index: _KeyIndexType,
        entities: tuple[type[Any], ...] = (),
    ):
        self._real_result = real_result

        if real_result._source_supports_scalars:
            self._metadata = real_result._metadata
            self._post_creational_filter = None
        else:
            self._metadata = real_result._metadata._reduce([index])
            self._post_creational_filter = operator.itemgetter(0)

        self._unique_filter_state = real_result._unique_filter_state
        self._entities = entities

    @cached_property
    def scalar_adapter(self) -> TypeAdapter:
        """Get the adapter for scalar results."""
        return (
            get_cached_adapter(self._entities[0])
            if self._entities
            else TypeAdapter(object)
        )

    def _adapt_scalar(self, item: Any) -> Any:
        """Adapt a scalar result using TypeAdapter validation."""
        return self.scalar_adapter.validate_python(item)

    def unique(
        self,
        strategy: Optional[_UniqueFilterType] = None,
    ) -> Self:
        """Apply unique filtering to the objects returned by this result."""
        self._unique_filter_state = (set(), strategy)
        return self

    async def partitions(self, size: int | None = None) -> AsyncIterator[Sequence[_R]]:
        """Iterate through sub-lists of elements of the size given."""
        getter = self._manyrow_getter

        while True:
            partition = await greenlet_spawn(getter, self, size)
            if partition:
                yield [self._adapt_scalar(scalar) for scalar in partition]
            else:
                break

    async def fetchall(self) -> Sequence[_R]:
        """A synonym for the :meth:`AsyncAdaptedScalarResult.all` method."""
        scalars = await greenlet_spawn(self._allrows)
        return [self._adapt_scalar(scalar) for scalar in scalars]

    async def fetchmany(self, size: int | None = None) -> Sequence[_R]:
        """Fetch many objects."""
        scalars = await greenlet_spawn(self._manyrow_getter, self, size)
        return [self._adapt_scalar(scalar) for scalar in scalars]

    async def all(self) -> Sequence[_R]:
        """Return all scalar values in a list."""
        scalars = await greenlet_spawn(self._allrows)
        return [self._adapt_scalar(scalar) for scalar in scalars]

    def __aiter__(self) -> AsyncAdaptedScalarResult[_R]:
        return self

    async def __anext__(self) -> _R:
        row = await greenlet_spawn(self._onerow_getter, self)
        if row is _NO_ROW:
            raise StopAsyncIteration()
        else:
            return self._adapt_scalar(row)

    async def first(self) -> Optional[_R]:
        """Fetch the first object or ``None`` if no object is present."""
        scalar = await greenlet_spawn(self._only_one_row, False, False, False)
        return self._adapt_scalar(scalar) if scalar else None

    async def one_or_none(self) -> Optional[_R]:
        """Return at most one object or raise an exception."""
        scalar = await greenlet_spawn(self._only_one_row, True, False, False)
        return self._adapt_scalar(scalar) if scalar else None

    async def one(self) -> _R:
        """Return exactly one object or raise an exception."""
        scalar = await greenlet_spawn(self._only_one_row, True, True, False)
        return self._adapt_scalar(scalar)


class AsyncAdaptedMappingResult(_WithKeys, AsyncAdaptedCommon[RowMapping]):
    """An async wrapper that returns adapted dictionary values."""

    __slots__ = ()

    _generate_rows = True
    _post_creational_filter = operator.attrgetter("_mapping")
    _entities: tuple[type[Any], ...]

    def __init__(
        self,
        result: Result[Any],
        entities: tuple[type[Any], ...] = (),
    ):
        self._real_result = result
        self._metadata = result._metadata
        self._unique_filter_state = result._unique_filter_state
        if result._source_supports_scalars:
            self._metadata = self._metadata._reduce([0])

        self._entities = entities

    @cached_property
    def adapter(self) -> TypeAdapter:
        """Get the adapter for tuple results."""
        return get_cached_adapter(tuple[*self._entities])

    def _adapt(self, row: Any) -> Any:
        """Adapt a row using TypeAdapter validation."""
        return self.adapter.validate_python(row)

    @_generative
    def unique(self, strategy: Optional[_UniqueFilterType] = None) -> Self:
        self._unique_filter_state = (set(), strategy)
        return self

    def columns(self, *col_expressions: _KeyIndexType) -> Self:
        return self._column_slices(col_expressions)

    async def partitions(
        self, size: int | None = None
    ) -> AsyncIterator[Sequence[dict[str, Any]]]:
        """Iterate through sub-lists of elements of the size given."""
        getter = self._manyrow_getter

        while True:
            partition = await greenlet_spawn(getter, self, size)
            if partition:
                yield list(
                    dict(zip(row.keys(), self._adapt(row.values())))
                    for row in partition
                )
            else:
                break

    async def fetchone(self) -> Optional[dict[str, Any]]:
        """Fetch one object."""
        row = await greenlet_spawn(self._onerow_getter, self)
        if row is _NO_ROW:
            return None
        else:
            return dict(zip(row.keys(), self._adapt(row.values())))

    async def fetchmany(self, size: int | None = None) -> Sequence[dict[str, Any]]:
        """Fetch many rows."""
        rows = await greenlet_spawn(self._manyrow_getter, self, size)
        return list(dict(zip(row.keys(), self._adapt(row.values()))) for row in rows)

    async def fetchall(self) -> Sequence[dict[str, Any]]:
        """A synonym for the :meth:`AsyncAdaptedMappingResult.all` method."""
        rows = await greenlet_spawn(self._allrows)
        return list(dict(zip(row.keys(), self._adapt(row.values()))) for row in rows)

    async def all(self) -> Sequence[dict[str, Any]]:
        """Return all rows in a list."""
        rows = await greenlet_spawn(self._allrows)
        return list(dict(zip(row.keys(), self._adapt(row.values()))) for row in rows)

    def __aiter__(self) -> AsyncAdaptedMappingResult:
        return self

    async def __anext__(self) -> dict[str, Any]:
        row = await greenlet_spawn(self._onerow_getter, self)
        if row is _NO_ROW:
            raise StopAsyncIteration()
        else:
            return dict(zip(row.keys(), self._adapt(row.values())))

    async def first(self) -> Optional[dict[str, Any]]:
        """Fetch the first object or ``None`` if no object is present."""
        row = await greenlet_spawn(self._only_one_row, False, False, False)
        if row:
            return dict(zip(row.keys(), self._adapt(row.values())))
        return None

    async def one_or_none(self) -> Optional[dict[str, Any]]:
        """Return at most one object or raise an exception."""
        row = await greenlet_spawn(self._only_one_row, True, False, False)
        if row:
            return dict(zip(row.keys(), self._adapt(row.values())))
        return None

    async def one(self) -> dict[str, Any]:
        """Return exactly one object or raise an exception."""
        row = await greenlet_spawn(self._only_one_row, True, True, False)
        return dict(zip(row.keys(), self._adapt(row.values())))



"""Test model_construct with SQLAlchemy materia.

This module tests:
- model_construct with SQLAlchemy ORM instances (provider data)
- model_construct with circular references
- model_construct validation context handling
- model_construct with transmuter_proxy relationships
- model_construct with materia hooks (before_construct/after_construct)
"""

from __future__ import annotations

from sqlalchemy import Engine

from arcanus.base import validation_context
from arcanus.materia.sqlalchemy import Session
from tests import models
from tests.transmuters import Author, Book


class TestModelConstructWithProvider:
    """Test model_construct with SQLAlchemy provider instances."""

    def test_construct_from_orm_instance(self, engine: Engine):
        """Test constructing from an ORM instance."""
        with Session(engine) as session:
            with session.begin():
                # Create ORM instance
                orm_author = models.Author(name="Construct Author", field="Physics")
                session.add(orm_author)
                session.flush()

                # Construct transmuter from ORM instance
                author = Author.model_construct(data=orm_author)

                assert author.name == "Construct Author"
                assert author.field == "Physics"
                assert author.__transmuter_provided__ is orm_author
                assert orm_author.transmuter_proxy is author

    def test_construct_caches_in_context(self, engine: Engine):
        """Test that constructed instances are cached in validation context."""
        with Session(engine) as session:
            with session.begin():
                orm_author = models.Author(name="Cached Author", field="Biology")
                session.add(orm_author)
                session.flush()

                with validation_context() as ctx:
                    # First construction
                    author1 = Author.model_construct(data=orm_author)

                    # Second construction of same ORM instance
                    author2 = Author.model_construct(data=orm_author)

                    # Should return the same cached instance
                    assert author1 is author2
                    assert ctx[orm_author] is author1

    def test_construct_respects_existing_proxy(self, engine: Engine):
        """Test that existing transmuter_proxy is respected."""
        with Session(engine) as session:
            with session.begin():
                orm_author = models.Author(name="Proxy Author", field="Chemistry")
                session.add(orm_author)
                session.flush()

                # First create through validation
                author1 = Author.model_validate(orm_author)

                # Now construct - should use existing proxy
                with validation_context():
                    author2 = Author.model_construct(data=orm_author)

                    # Should use the existing proxy
                    assert author2 is author1

    def test_construct_transmuter_proxy(self, engine: Engine):
        """Test constructing when proxy is None."""
        with Session(engine) as session:
            with session.begin():
                orm_author = models.Author(name="New Author", field="History")
                session.add(orm_author)
                session.flush()

                # Construct should create new instance
                author = Author.model_construct(data=orm_author)

                assert author.name == "New Author"
                assert orm_author.transmuter_proxy is author


class TestModelConstructCircularReferences:
    """Test model_construct with circular references."""

    def test_construct_with_bidirectional_relationship(self, engine: Engine):
        """Test constructing objects with bidirectional relationships."""
        with Session(engine) as session:
            with session.begin():
                orm_publisher = models.Publisher(name="Test Publisher", country="USA")
                orm_author = models.Author(name="Circular Author", field="Physics")
                orm_book = models.Book(
                    title="Circular Book",
                    year=2024,
                    author=orm_author,
                    publisher=orm_publisher,
                )
                session.add_all([orm_publisher, orm_author, orm_book])
                session.flush()

                with validation_context() as ctx:
                    # Construct author (which has books relationship)
                    author = Author.model_construct(data=orm_author)

                    # Construct book (which references the same author)
                    book = Book.model_construct(data=orm_book)

                    # Both should be in context
                    assert ctx[orm_author] is author
                    assert ctx[orm_book] is book

    def test_construct_prevents_infinite_recursion(self, engine: Engine):
        """Test that circular references don't cause infinite recursion."""
        with Session(engine) as session:
            with session.begin():
                orm_publisher = models.Publisher(name="Test Publisher", country="USA")
                orm_author = models.Author(name="Recursive Author", field="Physics")
                orm_book1 = models.Book(
                    title="Book 1",
                    year=2024,
                    author=orm_author,
                    publisher=orm_publisher,
                )
                orm_book2 = models.Book(
                    title="Book 2",
                    year=2024,
                    author=orm_author,
                    publisher=orm_publisher,
                )
                session.add_all([orm_publisher, orm_author, orm_book1, orm_book2])
                session.flush()

                with validation_context():
                    # This should not cause infinite recursion
                    author = Author.model_construct(data=orm_author)

                    assert author.name == "Recursive Author"
                    # Association should be prepared
                    assert hasattr(author, "books")


class TestModelConstructMateriaHooks:
    """Test materia hooks during model_construct."""

    def test_construct_calls_before_construct(self, engine: Engine):
        """Test that transmuter_before_construct is called."""
        with Session(engine) as session:
            with session.begin():
                orm_author = models.Author(name="Hook Author", field="Physics")
                session.add(orm_author)
                session.flush()

                # The before_construct hook should extract data from ORM
                author = Author.model_construct(data=orm_author)

                # Should have extracted the data properly
                assert author.name == "Hook Author"
                assert author.field == "Physics"
                assert author.id is not None

    def test_construct_calls_after_construct(self, engine: Engine):
        """Test that transmuter_after_construct is called."""
        with Session(engine) as session:
            with session.begin():
                orm_book = models.Book(
                    title="Hook Book",
                    year=2024,
                    author=models.Author(name="Author", field="Physics"),
                    publisher=models.Publisher(name="Publisher", country="USA"),
                )
                session.add(orm_book)
                session.flush()

                # The after_construct hook should process the instance
                book = Book.model_construct(data=orm_book)

                assert book.title == "Hook Book"
                assert book.year == 2024


class TestModelConstructWithAssociations:
    """Test model_construct with association fields."""

    def test_construct_separates_association_fields(self, engine: Engine):
        """Test that association fields are properly separated."""
        with Session(engine) as session:
            with session.begin():
                orm_publisher = models.Publisher(name="Test Publisher", country="USA")
                orm_author = models.Author(name="Associated Author", field="Physics")
                orm_book = models.Book(
                    title="Associated Book",
                    year=2024,
                    author=orm_author,
                    publisher=orm_publisher,
                )
                session.add_all([orm_publisher, orm_author, orm_book])
                session.flush()

                # Construct with associations
                book = Book.model_construct(data=orm_book)

                # Regular fields should be set
                assert book.title == "Associated Book"
                assert book.year == 2024

                # Association should be prepared
                assert hasattr(book, "author")

    def test_construct_prepares_associations(self, engine: Engine):
        """Test that associations are properly prepared after construction."""
        with Session(engine) as session:
            with session.begin():
                orm_author = models.Author(name="Prepared Author", field="Biology")
                session.add(orm_author)
                session.flush()

                author = Author.model_construct(data=orm_author)

                # Accessing association should work
                books = author.books
                assert books is not None


class TestModelConstructMixedUsage:
    """Test mixed usage of model_construct."""

    def test_construct_then_validate(self, engine: Engine):
        """Test constructing then validating."""
        with Session(engine) as session:
            with session.begin():
                orm_author = models.Author(name="Mixed Author", field="Chemistry")
                session.add(orm_author)
                session.flush()

                # Construct without validation
                author = Author.model_construct(data=orm_author)

                # Now validate
                author.revalidate()

                assert author.name == "Mixed Author"
                assert author.field == "Chemistry"

    def test_construct_with_kwargs_overrides(self, engine: Engine):
        """Test that kwargs can override provider data."""
        with Session(engine) as session:
            with session.begin():
                orm_author = models.Author(name="Original", field="Physics")
                session.add(orm_author)
                session.flush()

                # Construct with overrides
                author = Author.model_construct(
                    data=orm_author,
                    name="Overridden",
                )

                # The override should take effect
                assert author.name == "Overridden"
                assert author.field == "Physics"

    def test_construct_with_provider(self, engine: Engine):
        """Test normal construction with provider data."""
        author = Author.model_construct(
            id=999,
            name="No Provider",
            field="Literature",
        )

        assert author.__transmuter_provided__
        assert isinstance(author.__transmuter_provided__, models.Author)
        assert author.id == author.__transmuter_provided__.id == 999
        assert author.name == author.__transmuter_provided__.name == "No Provider"
        assert author.field == author.__transmuter_provided__.field == "Literature"


class TestModelConstructEdgeCases:
    """Test edge cases for model_construct."""

    def test_construct_with_empty_data(self):
        """Test constructing with no data."""
        book = Book.model_construct()

        # Should create an instance with undefined fields
        assert isinstance(book, Book)

    def test_construct_with_none_data(self):
        """Test constructing with None as data."""
        book = Book.model_construct(data=None, title="Test", year=2024, author_id=1)

        assert book.title == "Test"
        assert book.year == 2024



"""Test RelationMap association with SQLAlchemy Materia.

Tests:
- CRUD operations with dict-based relationships (Shelf <-> ShelfItem via attribute_keyed_dict)
- Lazy and eager loading of dict relationships
- Serialization after ORM load
- Dict mutation operations persisted through SQLAlchemy
"""

from __future__ import annotations

from sqlalchemy import Engine, select
from sqlalchemy.orm import selectinload

from arcanus.association import RelationMap
from arcanus.materia.sqlalchemy import Session
from tests.transmuters import Shelf, ShelfItem


class TestRelationMapCRUD:
    """Test basic CRUD operations with dict-based relationships."""

    def test_create_shelf_with_items(self, engine: Engine):
        """Test creating a Shelf with ShelfItems through RelationMap."""
        with Session(engine) as session:
            shelf = Shelf(name="Science")
            item1 = ShelfItem(label="physics", description="Physics books")
            item2 = ShelfItem(label="chemistry", description="Chemistry books")
            shelf.items["physics"] = item1
            shelf.items["chemistry"] = item2

            session.add(shelf)
            session.flush()
            shelf.revalidate()
            item1.revalidate()
            item2.revalidate()

            assert shelf.id is not None
            assert item1.id is not None
            assert item2.id is not None
            assert len(shelf.items) == 2

    def test_create_shelf_empty(self, engine: Engine):
        """Test creating a Shelf with no items."""
        with Session(engine) as session:
            shelf = Shelf(name="Empty Shelf")
            session.add(shelf)
            session.flush()
            shelf.revalidate()

            assert shelf.id is not None
            assert len(shelf.items) == 0

    def test_add_item_to_existing_shelf(self, engine: Engine):
        """Test adding a ShelfItem to an existing Shelf."""
        with Session(engine) as session:
            shelf = Shelf(name="Growing Shelf")
            session.add(shelf)
            session.flush()

            item = ShelfItem(label="math", description="Math books")
            shelf.items["math"] = item
            session.flush()
            item.revalidate()

            assert item.id is not None
            assert len(shelf.items) == 1
            assert shelf.items["math"].description == "Math books"

    def test_remove_item_from_shelf(self, engine: Engine):
        """Test removing a ShelfItem from a Shelf."""
        with Session(engine) as session:
            shelf = Shelf(name="Shrinking Shelf")
            item = ShelfItem(label="temp", description="Temporary item")
            shelf.items["temp"] = item
            session.add(shelf)
            session.flush()

            assert len(shelf.items) == 1
            del shelf.items["temp"]
            session.flush()

            assert len(shelf.items) == 0

    def test_update_item_in_shelf(self, engine: Engine):
        """Test replacing a ShelfItem at an existing key."""
        with Session(engine) as session:
            shelf = Shelf(name="Update Shelf")
            item1 = ShelfItem(label="slot", description="Original")
            shelf.items["slot"] = item1
            session.add(shelf)
            session.flush()

            item2 = ShelfItem(label="slot", description="Replacement")
            shelf.items["slot"] = item2
            session.flush()

            assert shelf.items["slot"].description == "Replacement"

    def test_clear_all_items(self, engine: Engine):
        """Test clearing all items from a shelf."""
        with Session(engine) as session:
            shelf = Shelf(name="Clearable Shelf")
            for i in range(3):
                shelf.items[f"item{i}"] = ShelfItem(
                    label=f"item{i}", description=f"Item {i}"
                )
            session.add(shelf)
            session.flush()

            assert len(shelf.items) == 3
            shelf.items.clear()
            session.flush()

            assert len(shelf.items) == 0


class TestRelationMapLoading:
    """Test lazy and eager loading of dict-based relationships."""

    def test_lazy_load_items(self, engine: Engine):
        """Test that items are lazily loaded when accessed."""
        with Session(engine) as session:
            shelf = Shelf(name="Lazy Shelf")
            shelf.items["a"] = ShelfItem(label="a", description="Alpha")
            shelf.items["b"] = ShelfItem(label="b", description="Beta")
            session.add(shelf)
            session.flush()
            shelf.revalidate()
            shelf_id = shelf.id
            session.commit()

        # New session to test lazy loading
        with Session(engine) as session:
            stmt = select(Shelf).where(Shelf["id"] == shelf_id)
            shelf = session.scalars(stmt).unique().one()
            assert isinstance(shelf, Shelf)
            assert isinstance(shelf.items, RelationMap)

            # Accessing items triggers lazy load
            assert len(shelf.items) == 2
            assert shelf.items["a"].description == "Alpha"
            assert shelf.items["b"].description == "Beta"

    def test_eager_load_selectinload(self, engine: Engine):
        """Test eager loading items with selectinload."""
        with Session(engine) as session:
            shelf = Shelf(name="Eager Shelf")
            shelf.items["x"] = ShelfItem(label="x", description="X-ray")
            shelf.items["y"] = ShelfItem(label="y", description="Yankee")
            session.add(shelf)
            session.flush()
            shelf.revalidate()
            shelf_id = shelf.id
            session.commit()

        with Session(engine) as session:
            stmt = (
                select(Shelf)
                .where(Shelf["id"] == shelf_id)
                .options(selectinload(Shelf["items"]))
            )
            shelf = session.scalars(stmt).unique().one()
            assert isinstance(shelf, Shelf)
            assert len(shelf.items) == 2

    def test_load_preserves_keys(self, engine: Engine):
        """Test that dict keys are preserved after loading from DB."""
        with Session(engine) as session:
            shelf = Shelf(name="Key Shelf")
            shelf.items["alpha"] = ShelfItem(label="alpha", description="First")
            shelf.items["beta"] = ShelfItem(label="beta", description="Second")
            shelf.items["gamma"] = ShelfItem(label="gamma", description="Third")
            session.add(shelf)
            session.flush()
            shelf.revalidate()
            shelf_id = shelf.id
            session.commit()

        with Session(engine) as session:
            stmt = select(Shelf).where(Shelf["id"] == shelf_id)
            shelf = session.scalars(stmt).unique().one()
            assert set(shelf.items.keys()) == {"alpha", "beta", "gamma"}


class TestRelationMapSerialization:
    """Test serialization of RelationMap after ORM operations."""

    def test_model_dump_after_create(self, engine: Engine):
        """Test model_dump after creating shelf with items."""
        with Session(engine) as session:
            shelf = Shelf(name="Dump Shelf")
            shelf.items["py"] = ShelfItem(label="py", description="Python")
            shelf.items["rs"] = ShelfItem(label="rs", description="Rust")
            session.add(shelf)
            session.flush()

            data = shelf.model_dump()
            assert isinstance(data["items"], dict)
            assert set(data["items"].keys()) == {"py", "rs"}
            assert data["items"]["py"]["description"] == "Python"

    def test_model_dump_after_load(self, engine: Engine):
        """Test model_dump after loading from DB with selectinload."""
        with Session(engine) as session:
            shelf = Shelf(name="Load Dump Shelf")
            shelf.items["a"] = ShelfItem(label="a", description="Alpha")
            session.add(shelf)
            session.flush()
            shelf.revalidate()
            shelf_id = shelf.id
            session.commit()

        with Session(engine) as session:
            stmt = (
                select(Shelf)
                .where(Shelf["id"] == shelf_id)
                .options(selectinload(Shelf["items"]))
            )
            shelf = session.scalars(stmt).unique().one()
            data = shelf.model_dump()
            assert isinstance(data["items"], dict)
            assert "a" in data["items"]
            assert data["items"]["a"]["description"] == "Alpha"

    def test_model_dump_empty_items(self, engine: Engine):
        """Test model_dump with empty items dict."""
        with Session(engine) as session:
            shelf = Shelf(name="Empty Dump Shelf")
            session.add(shelf)
            session.flush()

            data = shelf.model_dump()
            assert data["items"] == {}


class TestRelationMapDictOperations:
    """Test dict operations through the ORM."""

    def test_get_existing_key(self, engine: Engine):
        """Test get() on a loaded shelf."""
        with Session(engine) as session:
            shelf = Shelf(name="Get Shelf")
            shelf.items["k"] = ShelfItem(label="k", description="Known")
            session.add(shelf)
            session.flush()

            assert shelf.items.get("k") is not None
            assert shelf.items["k"].description == "Known"
            assert shelf.items.get("missing") is None

    def test_pop_from_shelf(self, engine: Engine):
        """Test pop() removes and returns the item."""
        with Session(engine) as session:
            shelf = Shelf(name="Pop Shelf")
            shelf.items["p"] = ShelfItem(label="p", description="Poppable")
            session.add(shelf)
            session.flush()

            item = shelf.items.pop("p")
            assert item.description == "Poppable"
            assert len(shelf.items) == 0

    def test_keys_values_items(self, engine: Engine):
        """Test keys(), values(), items() views."""
        with Session(engine) as session:
            shelf = Shelf(name="Views Shelf")
            shelf.items["one"] = ShelfItem(label="one", description="First")
            shelf.items["two"] = ShelfItem(label="two", description="Second")
            session.add(shelf)
            session.flush()

            assert set(shelf.items.keys()) == {"one", "two"}
            assert len(list(shelf.items.values())) == 2
            assert len(list(shelf.items.items())) == 2

    def test_contains_key(self, engine: Engine):
        """Test key membership check."""
        with Session(engine) as session:
            shelf = Shelf(name="Contains Shelf")
            shelf.items["here"] = ShelfItem(label="here", description="Present")
            session.add(shelf)
            session.flush()

            assert "here" in shelf.items
            assert "gone" not in shelf.items

    def test_update_dict(self, engine: Engine):
        """Test update() with a dict of new items."""
        with Session(engine) as session:
            shelf = Shelf(name="Update Shelf")
            session.add(shelf)
            session.flush()

            shelf.items.update(
                {
                    "a": ShelfItem(label="a", description="Alpha"),
                    "b": ShelfItem(label="b", description="Beta"),
                }
            )
            session.flush()

            assert len(shelf.items) == 2
            assert shelf.items["a"].description == "Alpha"

    def test_setdefault(self, engine: Engine):
        """Test setdefault() inserts only if key is absent."""
        with Session(engine) as session:
            shelf = Shelf(name="Setdefault Shelf")
            existing = ShelfItem(label="k", description="Original")
            shelf.items["k"] = existing
            session.add(shelf)
            session.flush()

            new = ShelfItem(label="k", description="New")
            result = shelf.items.setdefault("k", new)
            assert result.description == "Original"


class TestRelationMapBackReference:
    """Test back-reference from ShelfItem to Shelf."""

    def test_item_has_shelf_reference(self, engine: Engine):
        """Test that ShelfItem can reference its parent Shelf."""
        with Session(engine) as session:
            shelf = Shelf(name="BackRef Shelf")
            item = ShelfItem(label="ref", description="Has back-ref")
            shelf.items["ref"] = item
            session.add(shelf)
            session.flush()

            assert item.shelf.value is not None
            assert item.shelf.value.name == "BackRef Shelf"



from arcanus.materia.sqlalchemy.base import SqlalchemyMateria
from arcanus.materia.sqlalchemy.database import AsyncSession, Session

__all__ = [
    "SqlalchemyMateria",
    "Session",
    "AsyncSession",
]



from __future__ import annotations

from functools import cached_property, partial, wraps
from types import UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Concatenate,
    ForwardRef,
    Generic,
    Iterable,
    Literal,
    Mapping,
    Optional,
    ParamSpec,
    Self,
    SupportsIndex,
    Type,
    TypeVar,
    Union,
    cast,
    final,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from pydantic import Field, GetCoreSchemaHandler, TypeAdapter
from pydantic_core import core_schema
from typing_extensions import is_typeddict

from arcanus.materia.base import active_materia
from arcanus.utils import get_cached_adapter

if TYPE_CHECKING:
    from _typeshed import SupportsRichComparison

    from arcanus.base import Transmuter

A = TypeVar("A")
D = TypeVar("D")
K = TypeVar("K")
T = TypeVar("T", bound="Transmuter")
TD = TypeVar("TD")  # TypedDict type parameter for TypedRelationMap
Optional_T = TypeVar("Optional_T", bound="Transmuter | Optional[Transmuter]")

P = ParamSpec("P")
R = TypeVar("R")


@final
class DefferedAssociation:
    """A type used as a sentinel for already loaded association values, for deffering the blessing"""

    def __copy__(self) -> Self: ...


def is_association(t: type) -> bool:
    origin = get_origin(t)
    arg = origin or t

    # Union of scalar types, e.g. Union[int, str] or Optional[str], which is Union[str, NoneType]
    if origin is Union or origin is UnionType:
        arg = get_args(t)[0]

    # Literal types, e.g. Literal["value1", "value2"]
    if origin is Literal:
        arg = type(get_args(t)[0])

    return issubclass(arg, Association)


class Association(Generic[A]):
    __args__: tuple[type, ...]
    __instance__: Transmuter | None
    __loaded__: bool
    __payloads__: A | None

    field_name: str

    @classmethod
    def __get_pydantic_generic_schema__(
        cls,
        generic_type: Type[A],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        raise NotImplementedError()

    @classmethod
    def __get_pydantic_serialize_schema__(
        cls,
        generic_type: Type[A],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.SerSchema | None:
        """
        The default serialization schema for associations is to serialize the loaded value if the association is loaded, otherwise serialize the payloads.
        """
        # TODO: Implement automatic circular reference detection in serialization.
        # Currently, circular references must be manually excluded using the exclude
        # parameter. Pydantic does not provide built-in cycle detection.
        # See: https://docs.pydantic.dev/latest/concepts/serialization/
        raise NotImplementedError()

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Type[Association[A]], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        args = get_args(source_type)

        if not args:
            raise TypeError(f"Generic type must be provided to the {source_type}.")

        generic_type = args[0]

        def validate(
            value: Any,
            handler: core_schema.ValidatorFunctionWrapHandler,
            info: core_schema.ValidationInfo,
        ) -> Association[A]:
            # if not info.field_name:
            #     raise ValueError(
            #         f"The association type {source_type} must be used as a model field."
            #     )

            # materia = active_materia.get()
            # value = materia.association_before_validator(cls, value, info)

            if value is DefferedAssociation:
                instance = cls()
            elif type(value) is cls:
                instance = value
                instance.__payloads__ = handler(instance.__payloads__)
            else:
                instance = cls(handler(value))

            instance.__args__ = (generic_type,)
            instance.field_name = info.field_name  # pyright: ignore[reportAttributeAccessIssue]
            # instance = materia.association_after_validator(instance, info)

            return instance

        return core_schema.with_default_schema(
            core_schema.with_info_wrap_validator_function(
                validate,
                cls.__get_pydantic_generic_schema__(generic_type, handler),
            ),
            default_factory=cls,
            serialization=cls.__get_pydantic_serialize_schema__(generic_type, handler),
        )

    @property
    def __instance_provider__(self) -> Optional[Any]:
        """Owner instance' provider, owner of this association's provider."""
        if self.__instance__ is not None:
            return self.__instance__.__transmuter_provided__
        return None

    @property
    def __provided__(self) -> Any | None:
        raise NotImplementedError()

    @cached_property
    def __validator__(self) -> TypeAdapter[A]:
        return get_cached_adapter(self.__args__[0])

    @cached_property
    def used_name(self) -> str:
        return (
            alias
            if self.__instance__ and (alias := self.field_info.alias)
            else self.field_name
        )

    def __init__(self, payloads: A | None = None):
        self.__instance__ = None
        self.__loaded__ = False
        self.__payloads__ = payloads

    def __await__(self):
        raise NotImplementedError("This association is not awaitable.")

    def _load(self) -> Self:
        raise NotImplementedError(
            "This association does not support synchronous loading."
        )

    def _aload(self) -> Self:
        raise NotImplementedError(
            "This association does not support asynchronous loading."
        )

    def prepare(self, instance: Transmuter, field_name: str):
        if self.__instance__ is not None:
            return

        self.field_name = field_name
        self.field_info = type(instance).__pydantic_fields__[field_name]

        self.__instance__ = instance

        annotation = self.field_info.annotation
        if isinstance(annotation, ForwardRef):
            resolved_hints = get_type_hints(type(instance))
            actual_type = resolved_hints[field_name]
            self.__args__ = (get_args(actual_type)[0],)
        else:
            self.__args__ = (get_args(annotation)[0],)

    def bless(self, value: Any) -> Any:
        """Bless the value into the generic type."""
        return self.__validator__.validate_python(value)


class Relation(Association[Optional_T]):
    # new item and loaded item are shared the __payloads__ here
    __payloads__: Optional_T

    @classmethod
    def __get_pydantic_generic_schema__(
        cls, generic_type: type[Optional_T], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        # TODO: strict the validation for lazy-load non-optional single relationship
        # to fobid the folowing example
        # class A(BaseTransmuter):
        #     b: Relation[B] = Relation()
        # a = A(b=None)  # should raise validation error
        return core_schema.union_schema(
            choices=[
                handler.generate_schema(generic_type),
                core_schema.none_schema(),
            ]
        )

        # return handler.generate_schema(generic_type)

    @classmethod
    def __get_pydantic_serialize_schema__(
        cls, generic_type: type[Optional_T], handler: GetCoreSchemaHandler
    ) -> core_schema.SerSchema | None:
        def serialize(association: Relation[Optional_T], serializer) -> Any:
            fields_set = getattr(
                association.__instance__, "__pydantic_fields_set__", None
            )
            if fields_set is not None and association.field_name in fields_set:
                return serializer(association.value)
            return serializer(association.__payloads__)

        return core_schema.wrap_serializer_function_ser_schema(
            serialize,
            schema=handler.generate_schema(generic_type),
            when_used="always",
        )

    @property
    def __provided__(self) -> Any:
        if not self.__instance_provider__:
            return None

        # TODO: provider not exist, or the provided value is None both return None
        return getattr(self.__instance_provider__, self.used_name)

    @__provided__.setter
    def __provided__(self, object: Any):
        if not self.__instance_provider__:
            return  # No provider, skip syncing
        setattr(self.__instance_provider__, self.used_name, object)

    def prepare(self, instance: Transmuter, field_name: str):
        super().prepare(instance, field_name)
        if self.__payloads__ is not None:
            self._load()
            self.__provided__ = self.__payloads__.__transmuter_provided__

    @staticmethod
    def ensure_loaded(
        func: Callable[Concatenate[Relation[Optional_T], P], R],
    ) -> Callable[Concatenate[Relation[Optional_T], P], R]:
        @wraps(func)
        def wrapper(self: Relation[Optional_T], *args: P.args, **kwargs: P.kwargs) -> R:
            self._load()
            return func(self, *args, **kwargs)

        return wrapper

    def _load(self) -> Optional_T:
        # maybe during deepcopy from field default, or the relationship is already loaded
        if not self.__instance__ or self.__loaded__:
            return self.__payloads__

        active_materia.get().load_association(self)

        # A: No provided, None
        # B: provided value is None
        if not self.__provided__:
            return self.__payloads__

        if self.__payloads__ is not None and self.__payloads__.__transmuter_provided__:
            # Already loaded by ORM (e.g., selectinload), no need to set back
            self.__provided__ = self.__payloads__.__transmuter_provided__
        else:
            self.__payloads__ = self.bless(self.__provided__)

        self.__loaded__ = True

        return self.__payloads__

    async def _aload(self) -> Optional_T:
        # maybe during deepcopy from field default, or the relationship is already loaded
        if not self.__instance__ or self.__loaded__:
            return self.__payloads__

        await active_materia.get().aload_association(self)

        # A: No provided, None
        # B: provided value is None
        if not self.__provided__:
            return self.__payloads__

        if self.__payloads__ is not None and self.__payloads__.__transmuter_provided__:
            # Already loaded by ORM (e.g., selectinload), no need to set back
            self.__provided__ = self.__payloads__.__transmuter_provided__
        else:
            self.__payloads__ = self.bless(self.__provided__)

        self.__loaded__ = True

        return self.__payloads__

    def __await__(self):
        return self._aload().__await__()

    @property
    @ensure_loaded
    def value(self) -> Optional_T:
        return self.__payloads__

    @value.setter
    @ensure_loaded
    def value(self, object: Optional_T):
        object = self.bless(object)
        if object is not None:
            self.__provided__ = object.__transmuter_provided__
        else:
            self.__provided__ = None
        self.__payloads__ = object


# built-in types must be put at front to avoid pydantic convert it to built-in types
class RelationCollection(list[T], Association[T]):
    # new items are held in __payloads__, loaded items are kept in the list itself
    __payloads__: list[T]

    @classmethod
    def __get_pydantic_generic_schema__(
        cls,
        generic_type: Type[T],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        return core_schema.list_schema(handler.generate_schema(generic_type))

    @classmethod
    def __get_pydantic_serialize_schema__(
        cls, generic_type: Type[T], handler: GetCoreSchemaHandler
    ) -> core_schema.SerSchema | None:
        def serialize(association: RelationCollection[T], serializer) -> Any:
            fields_set = getattr(
                association.__instance__, "__pydantic_fields_set__", None
            )
            if fields_set is not None and association.field_name in fields_set:
                return serializer(association.copy())
            return serializer(list.copy(association) + association.__payloads__)

        return core_schema.wrap_serializer_function_ser_schema(
            serialize,
            schema=core_schema.list_schema(handler.generate_schema(generic_type)),
            when_used="always",
        )

    def __init__(self, payloads: Iterable[T] | None = None):
        super().__init__()
        self.__instance__ = None
        self.__loaded__ = False
        self.__payloads__ = list(payloads) if payloads else []

    @property
    def __provided__(self) -> list[Any] | None:
        # The return type should be a duck typed list-like object provided by the current materia provider.
        # For example, with SQLAlchemyMateria, it would be a InstrumentedList[list[...]] which is actually a sqlalchemy descriptor.
        if not self.__instance_provider__:
            return None
        return getattr(self.__instance_provider__, self.used_name)

    @cached_property
    def __list_validator__(self) -> TypeAdapter[list[T]]:
        return get_cached_adapter(list[self.__args__[0]])

    @overload
    def bless(self, value: T) -> T: ...
    @overload
    def bless(self, value: Iterable[Any]) -> list[T]: ...
    @overload
    def bless(self, value: Any) -> T: ...
    def bless(self, value: Any | Iterable[Any]) -> T | Iterable[T]:
        """Bless the value into the generic type."""
        is_iterable = isinstance(value, Iterable) and not isinstance(
            value, get_origin(self.__args__[0]) or self.__args__[0]
        )

        if is_iterable:
            return self.__list_validator__.validate_python(value)
        else:
            return self.__validator__.validate_python(value)

    def prepare(self, instance: Transmuter, field_name: str):
        super().prepare(instance, field_name)
        if self.__payloads__:
            # manualy enforce loading first to remove duplicates in payloads
            # objects already assigned to the relationship may be add to payloads during revalidation
            self._load()
            self.extend(self.__payloads__)
            self.__payloads__.clear()

    @staticmethod
    def ensure_loaded(
        func: Callable[Concatenate[RelationCollection[T], P], R],
    ) -> Callable[Concatenate[RelationCollection[T], P], R]:
        @wraps(func)
        def wrapper(
            self: RelationCollection[T], *args: P.args, **kwargs: P.kwargs
        ) -> R:
            self._load()
            return func(self, *args, **kwargs)

        return wrapper

    def _load(self):
        # maybe during deepcopy from field default
        if not self.__instance__:
            return self

        # or the relationship is already loaded
        if self.__loaded__:
            return self

        active_materia.get().load_association(self)

        # A: No provided, None
        # B: provided value is empty, []
        if not (self.__provided__):
            return self

        # TODO: Better way to avoid duplication relationship append ?
        self.__payloads__ = [
            payload
            for payload in self.__payloads__
            if payload.__transmuter_provided__ not in set(self.__provided__)
        ]

        if not len(self.__provided__) == super().__len__():
            # If the length of __provided__ is not equal to the length of self,
            # it means some items were not blessed into transmuter objects.
            super().clear()
            super().extend(self.bless(self.__provided__))
        self.__loaded__ = True

        return self

    async def _aload(self):
        # maybe during deepcopy from field default
        if not self.__instance__:
            return self

        # or the relationship is already loaded
        if self.__loaded__:
            return self

        # A: No provided, None
        # B: provided value is empty, []
        if not (provided := await active_materia.get().aload_association(self)):
            return self

        # TODO: Better way to avoid duplication relationship append ?
        self.__payloads__ = [
            payload
            for payload in self.__payloads__
            if payload.__transmuter_provided__ not in set(provided)
        ]

        if not len(provided) == super().__len__():
            # If the length of __provided__ is not equal to the length of self,
            # it means some items were not blessed into transmuter objects.
            super().clear()
            super().extend(self.bless(provided))
        self.__loaded__ = True

        return self

    def __await__(self):
        return self._aload().__await__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> T: ...
    @overload
    def __getitem__(self, index: slice) -> list[T]: ...
    @ensure_loaded
    def __getitem__(self, index: SupportsIndex | slice) -> T | list[T]:
        return super().__getitem__(index)

    @ensure_loaded
    def __iter__(self):
        return super().__iter__()

    @ensure_loaded
    def __len__(self):
        return super().__len__()

    @ensure_loaded
    def __contains__(self, key: T) -> bool:
        return super().__contains__(key)

    @ensure_loaded
    def __bool__(self):
        return super().__len__() > 0

    @overload
    def __setitem__(self, key: SupportsIndex, value: T) -> None: ...
    @overload
    def __setitem__(self, key: slice, value: Iterable[T]) -> None: ...
    @ensure_loaded
    def __setitem__(self, key: SupportsIndex | slice, value: T | Iterable[T]):
        if isinstance(value, Iterable):
            items = self.bless(value)
            slc = cast(slice, key)
            if self.__provided__ is not None:
                self.__provided__[slc] = [
                    item.__transmuter_provided__ for item in items
                ]
            super().__setitem__(slc, items)
        else:
            item = self.bless(value)
            idx = cast(SupportsIndex, key)
            if self.__provided__ is not None:
                self.__provided__[idx] = item.__transmuter_provided__
            super().__setitem__(idx, item)

    @ensure_loaded
    def __delitem__(self, key: slice):
        if self.__provided__ is not None:
            self.__provided__.__delitem__(key)
        super().__delitem__(key)

    @ensure_loaded
    def __add__(self, other: Iterable[T]):
        return self.copy() + self.bless(other)

    @ensure_loaded
    def __iadd__(self, other: Iterable[T]):
        self.extend(other)
        return self

    def __mul__(self, other):
        raise NotImplementedError(
            "Left multiplication on relationship is not supported."
        )

    def __rmul__(self, other):
        raise NotImplementedError(
            "Right multiplication on relationship is not supported."
        )

    def __imul__(self, other):
        raise NotImplementedError(
            "Self multiplication on relationship is not supported."
        )

    @ensure_loaded
    def __eq__(self, other: list[T]):
        return super().__eq__(other)

    @ensure_loaded
    def __ne__(self, other: list[T]):
        return super().__ne__(other)

    @ensure_loaded
    def __lt__(self, other: list[T]):
        return super().__lt__(other)

    @ensure_loaded
    def __le__(self, other: list[T]):
        return super().__le__(other)

    @ensure_loaded
    def __gt__(self, other: list[T]):
        return super().__gt__(other)

    @ensure_loaded
    def __ge__(self, other: list[T]):
        return super().__ge__(other)

    # @ensure_loaded
    def __repr__(self):
        # return super().__repr__()
        return f"RelationCollection[{self.__args__[0].__name__}], instance={id(self.__instance__)}, size={super().__len__()}"

    @ensure_loaded
    def __str__(self):
        return super().__str__()

    @ensure_loaded
    def __reversed__(self):
        return super().__reversed__()

    @ensure_loaded
    def append(self, value: T):
        value = self.bless(value)
        if self.__provided__ is not None:
            self.__provided__.append(value.__transmuter_provided__)
        super().append(value)

    @ensure_loaded
    def extend(self, iterable: Iterable[T]):
        iterable = self.bless(iterable)
        if self.__provided__ is not None:
            self.__provided__.extend(
                (item.__transmuter_provided__ for item in iterable)
            )
        super().extend(iterable)

    @ensure_loaded
    def clear(self):
        if self.__provided__ is not None:
            self.__provided__.clear()
        super().clear()

    @ensure_loaded
    def copy(self):
        return super().copy()

    @ensure_loaded
    def count(self, value: T) -> int:
        return super().count(value)

    @ensure_loaded
    def index(self, value, start=0, stop=None):
        if stop is None:
            return super().index(value, start)
        return super().index(value, start, stop)

    @ensure_loaded
    def insert(self, index: SupportsIndex, object: T):
        object = self.bless(object)
        if self.__provided__ is not None:
            self.__provided__.insert(index, object.__transmuter_provided__)
        super().insert(index, object)

    @ensure_loaded
    def pop(self, index: SupportsIndex = -1):
        item = super().pop(index)
        if self.__provided__ is not None:
            self.__provided__.remove(item.__transmuter_provided__)
        return item

    @ensure_loaded
    def remove(self, value: T):
        item: T = self.bless(value)
        if self.__provided__ is not None:
            self.__provided__.remove(item.__transmuter_provided__)
        super().remove(value)

    @ensure_loaded
    def reverse(self):
        super().reverse()

    @ensure_loaded
    def sort(
        self,
        *,
        key: Callable[[T], SupportsRichComparison],
        reverse: bool = False,
    ):
        super().sort(key=key, reverse=reverse)


# built-in types must be put at front to avoid pydantic convert it to built-in types
class RelationSet(set[T], Association[T]):
    # new items are held in __payloads__, loaded items are kept in the set itself
    __payloads__: set[T]

    @classmethod
    def __get_pydantic_generic_schema__(
        cls,
        generic_type: Type[T],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        return core_schema.set_schema(handler.generate_schema(generic_type))

    @classmethod
    def __get_pydantic_serialize_schema__(
        cls, generic_type: Type[T], handler: GetCoreSchemaHandler
    ) -> core_schema.SerSchema | None:
        def serialize(association: RelationSet[T], serializer) -> Any:
            fields_set = getattr(
                association.__instance__, "__pydantic_fields_set__", None
            )
            if fields_set is not None and association.field_name in fields_set:
                return serializer(association.copy())
            return serializer(list(set.copy(association) | association.__payloads__))

        return core_schema.wrap_serializer_function_ser_schema(
            serialize,
            schema=core_schema.list_schema(handler.generate_schema(generic_type)),
            when_used="always",
        )

    def __init__(self, payloads: Iterable[T] | None = None):
        super().__init__()
        self.__instance__ = None
        self.__loaded__ = False
        self.__payloads__ = set(payloads) if payloads else set()

    @property
    def __provided__(self) -> Any | None:
        # The return type should be a duck typed set-like object provided by the current materia provider.
        # For example, with SQLAlchemyMateria and collection_class=set, it would be an InstrumentedSet.
        if not self.__instance_provider__:
            return None
        return getattr(self.__instance_provider__, self.used_name)

    @cached_property
    def __set_validator__(self) -> TypeAdapter[set[T]]:
        return get_cached_adapter(set[self.__args__[0]])

    @overload
    def bless(self, value: T) -> T: ...
    @overload
    def bless(self, value: Iterable[Any]) -> set[T]: ...
    @overload
    def bless(self, value: Any) -> T: ...
    def bless(self, value: Any | Iterable[Any]) -> T | set[T]:
        """Bless the value into the generic type."""
        is_iterable = isinstance(value, Iterable) and not isinstance(
            value, get_origin(self.__args__[0]) or self.__args__[0]
        )

        if is_iterable:
            return self.__set_validator__.validate_python(value)
        else:
            return self.__validator__.validate_python(value)

    def prepare(self, instance: Transmuter, field_name: str):
        super().prepare(instance, field_name)
        if self.__payloads__:
            # manually enforce loading first to remove duplicates in payloads
            # objects already assigned to the relationship may be added to payloads during revalidation
            self._load()
            self.update(self.__payloads__)
            self.__payloads__.clear()

    @staticmethod
    def ensure_loaded(
        func: Callable[Concatenate[RelationSet[T], P], R],
    ) -> Callable[Concatenate[RelationSet[T], P], R]:
        @wraps(func)
        def wrapper(self: RelationSet[T], *args: P.args, **kwargs: P.kwargs) -> R:
            self._load()
            return func(self, *args, **kwargs)

        return wrapper

    def _load(self):
        # maybe during deepcopy from field default
        if not self.__instance__:
            return self

        # or the relationship is already loaded
        if self.__loaded__:
            return self

        active_materia.get().load_association(self)

        # A: No provided, None
        # B: provided value is empty
        if not self.__provided__:
            return self

        # Remove payloads that are already present in __provided__
        provided_set = set(self.__provided__)
        self.__payloads__ = {
            payload
            for payload in self.__payloads__
            if payload.__transmuter_provided__ not in provided_set
        }

        if len(self.__provided__) != super().__len__():
            # If the length of __provided__ is not equal to the length of self,
            # it means some items were not blessed into transmuter objects.
            super().clear()
            super().update(self.bless(self.__provided__))
        self.__loaded__ = True

        return self

    async def _aload(self):
        # maybe during deepcopy from field default
        if not self.__instance__:
            return self

        # or the relationship is already loaded
        if self.__loaded__:
            return self

        # A: No provided, None
        # B: provided value is empty
        if not (provided := await active_materia.get().aload_association(self)):
            return self

        # Remove payloads that are already present in provided
        provided_set = set(provided)
        self.__payloads__ = {
            payload
            for payload in self.__payloads__
            if payload.__transmuter_provided__ not in provided_set
        }

        if len(provided) != super().__len__():
            # If the length of __provided__ is not equal to the length of self,
            # it means some items were not blessed into transmuter objects.
            super().clear()
            super().update(self.bless(provided))
        self.__loaded__ = True

        return self

    def __await__(self):
        return self._aload().__await__()

    @ensure_loaded
    def __iter__(self):
        return super().__iter__()

    @ensure_loaded
    def __len__(self):
        return super().__len__()

    @ensure_loaded
    def __contains__(self, item: object) -> bool:
        return super().__contains__(item)

    @ensure_loaded
    def __bool__(self):
        return super().__len__() > 0

    def __repr__(self):
        return f"RelationSet[{self.__args__[0].__name__}], instance={id(self.__instance__)}, size={super().__len__()}"

    @ensure_loaded
    def __str__(self):
        return super().__str__()

    @ensure_loaded
    def add(self, item: T) -> None:
        """Add an element. No effect if already present (identity-based)."""
        item = self.bless(item)
        if item in self:
            return
        if self.__provided__ is not None:
            provided = item.__transmuter_provided__
            self.__provided__.add(provided)
        super().add(item)

    @ensure_loaded
    def discard(self, item: T) -> None:
        """Remove an element if present."""
        if item not in self:
            return
        if self.__provided__ is not None:
            self.__provided__.discard(item.__transmuter_provided__)
        super().discard(item)

    @ensure_loaded
    def remove(self, item: T) -> None:
        """Remove an element. Raises KeyError if not present."""
        if self.__provided__ is not None:
            self.__provided__.discard(item.__transmuter_provided__)
        super().remove(item)

    @ensure_loaded
    def pop(self) -> T:
        """Remove and return an arbitrary element. Raises KeyError if empty."""
        item = super().pop()
        if self.__provided__ is not None:
            self.__provided__.discard(item.__transmuter_provided__)
        return item

    @ensure_loaded
    def update(self, *others: Iterable[T]) -> None:
        """Add all elements from iterables."""
        for other in others:
            items = self.bless(other)
            for item in items:
                self.add(item)

    @ensure_loaded
    def clear(self) -> None:
        """Remove all elements."""
        if self.__provided__ is not None:
            self.__provided__.clear()
        super().clear()

    @ensure_loaded
    def intersection_update(self, *others: Iterable[T]) -> None:
        """Keep only elements found in all others."""
        keep = set.intersection(self, *others)
        removed = set.difference(self, keep)
        for item in removed:
            self.discard(item)

    @ensure_loaded
    def difference_update(self, *others: Iterable[T]) -> None:
        """Remove all elements found in others."""
        to_remove = set.intersection(self, *others)
        for item in to_remove:
            self.discard(item)

    @ensure_loaded
    def symmetric_difference_update(self, other: Iterable[T]) -> None:
        """Update to symmetric difference with other."""
        other_set = set(other)
        to_remove = set.intersection(self, other_set)
        to_add = other_set - set.copy(self)
        for item in to_remove:
            self.discard(item)
        for item in to_add:
            self.add(item)

    @ensure_loaded
    def copy(self) -> set[T]:
        return super().copy()

    @ensure_loaded
    def union(self, *others: Iterable[T]) -> set[T]:
        return super().union(*others)

    @ensure_loaded
    def intersection(self, *others: Iterable[T]) -> set[T]:
        return super().intersection(*others)

    @ensure_loaded
    def difference(self, *others: Iterable[T]) -> set[T]:
        return super().difference(*others)

    @ensure_loaded
    def symmetric_difference(self, other: Iterable[T]) -> set[T]:
        return super().symmetric_difference(other)

    @ensure_loaded
    def issubset(self, other: Iterable[T]) -> bool:
        return super().issubset(other)

    @ensure_loaded
    def issuperset(self, other: Iterable[T]) -> bool:
        return super().issuperset(other)

    @ensure_loaded
    def isdisjoint(self, other: Iterable[T]) -> bool:
        return super().isdisjoint(other)

    @ensure_loaded
    def __eq__(self, other: object) -> bool:
        if isinstance(other, RelationSet):
            return set.__eq__(self, other)
        if isinstance(other, (set, frozenset)):
            return set.__eq__(self, other)
        return False

    @ensure_loaded
    def __ne__(self, other: object) -> bool:
        if isinstance(other, RelationSet):
            return set.__ne__(self, other)
        if isinstance(other, (set, frozenset)):
            return set.__ne__(self, other)
        return True

    @ensure_loaded
    def __le__(self, other: set[T]) -> bool:
        return super().__le__(other)

    @ensure_loaded
    def __lt__(self, other: set[T]) -> bool:
        return super().__lt__(other)

    @ensure_loaded
    def __ge__(self, other: set[T]) -> bool:
        return super().__ge__(other)

    @ensure_loaded
    def __gt__(self, other: set[T]) -> bool:
        return super().__gt__(other)

    @ensure_loaded
    def __or__(self, other: set[T]) -> set[T]:
        return super().__or__(other)

    @ensure_loaded
    def __and__(self, other: set[T]) -> set[T]:
        return super().__and__(other)

    @ensure_loaded
    def __sub__(self, other: set[T]) -> set[T]:
        return super().__sub__(other)

    @ensure_loaded
    def __xor__(self, other: set[T]) -> set[T]:
        return super().__xor__(other)

    @ensure_loaded
    def __ior__(self, other: Iterable[T]) -> Self:
        self.update(other)
        return self

    @ensure_loaded
    def __iand__(self, other: Iterable[T]) -> Self:
        self.intersection_update(other)
        return self

    @ensure_loaded
    def __isub__(self, other: Iterable[T]) -> Self:
        self.difference_update(other)
        return self

    @ensure_loaded
    def __ixor__(self, other: Iterable[T]) -> Self:
        self.symmetric_difference_update(other)
        return self


# built-in types must be put at front to avoid pydantic convert it to built-in types
class RelationMap(dict[K, T], Association[T]):
    # new items are held in __payloads__, loaded items are kept in the dict itself
    # __args__[0] = key type (K), __args__[1] = value type (T)
    __payloads__: dict[K, T]

    @classmethod
    def __get_pydantic_generic_schema__(
        cls,
        key_type: Type[K],
        value_type: Type[T],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        return core_schema.dict_schema(
            keys_schema=handler.generate_schema(key_type),
            values_schema=handler.generate_schema(value_type),
        )

    @classmethod
    def __get_pydantic_serialize_schema__(
        cls,
        key_type: Type[K],
        value_type: Type[T],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.SerSchema | None:
        def serialize(association: RelationMap[K, T], serializer) -> Any:
            fields_set = getattr(
                association.__instance__, "__pydantic_fields_set__", None
            )
            if fields_set is not None and association.field_name in fields_set:
                return serializer(association.copy())
            return serializer(dict.copy(association) | association.__payloads__)

        return core_schema.wrap_serializer_function_ser_schema(
            serialize,
            schema=core_schema.dict_schema(
                keys_schema=handler.generate_schema(key_type),
                values_schema=handler.generate_schema(value_type),
            ),
            when_used="always",
        )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Type[RelationMap[K, T]], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        args = get_args(source_type)

        if not args or len(args) < 2:
            raise TypeError(
                f"Two generic types (key, value) must be provided to {source_type}."
            )

        key_type = args[0]
        value_type = args[1]

        def validate(
            value: Any,
            handler: core_schema.ValidatorFunctionWrapHandler,
            info: core_schema.ValidationInfo,
        ) -> RelationMap[K, T]:
            if value is DefferedAssociation:
                instance = cls()
            elif type(value) is cls:
                instance = value
                instance.__payloads__ = handler(instance.__payloads__)
            else:
                instance = cls(handler(value))

            instance.__args__ = (key_type, value_type)
            instance.field_name = info.field_name  # pyright: ignore[reportAttributeAccessIssue]

            return instance

        return core_schema.with_default_schema(
            core_schema.with_info_wrap_validator_function(
                validate,
                cls.__get_pydantic_generic_schema__(key_type, value_type, handler),
            ),
            default_factory=cls,
            serialization=cls.__get_pydantic_serialize_schema__(
                key_type, value_type, handler
            ),
        )

    def __init__(self, payloads: Mapping[K, T] | None = None):
        super().__init__()
        self.__instance__ = None
        self.__loaded__ = False
        self.__payloads__ = dict(payloads) if payloads else {}

    @property
    def __provided__(self) -> dict | None:
        # The return type should be a duck typed dict-like object provided by the current materia provider.
        # For example, with SQLAlchemyMateria and collection_class=attribute_keyed_dict,
        # it would be a KeyFuncDict.
        if not self.__instance_provider__:
            return None
        return getattr(self.__instance_provider__, self.used_name)

    @cached_property
    def __validator__(self) -> TypeAdapter[T]:
        return get_cached_adapter(self.__args__[1])

    @cached_property
    def __dict_validator__(self) -> TypeAdapter[dict[K, T]]:
        return get_cached_adapter(dict[self.__args__[0], self.__args__[1]])

    @cached_property
    def __key_validator__(self) -> TypeAdapter[K]:
        return get_cached_adapter(self.__args__[0])

    def bless_key(self, key: Any) -> K:
        """Validate and coerce a key into the key type."""
        return self.__key_validator__.validate_python(key)

    def bless_value(self, value: Any) -> T:
        """Validate and coerce a single value into the value type."""
        return self.__validator__.validate_python(value)

    def bless(self, value: Mapping[K, Any]) -> dict[K, T]:
        """Validate and coerce an entire dict/mapping into dict[K, T]."""
        return self.__dict_validator__.validate_python(value)

    def prepare(self, instance: Transmuter, field_name: str):
        if self.__instance__ is not None:
            return

        self.field_name = field_name
        self.field_info = type(instance).__pydantic_fields__[field_name]

        self.__instance__ = instance

        annotation = self.field_info.annotation
        if isinstance(annotation, ForwardRef):
            resolved_hints = get_type_hints(type(instance))
            actual_type = resolved_hints[field_name]
            args = get_args(actual_type)
        else:
            args = get_args(annotation)

        self.__args__ = (args[0], args[1])

        if self.__payloads__:
            # manually enforce loading first to remove duplicates in payloads
            # objects already assigned to the relationship may be added to payloads during revalidation
            self._load()
            self.update(self.__payloads__)
            self.__payloads__.clear()

    @staticmethod
    def ensure_loaded(
        func: Callable[Concatenate[RelationMap[K, T], P], R],
    ) -> Callable[Concatenate[RelationMap[K, T], P], R]:
        @wraps(func)
        def wrapper(self: RelationMap[K, T], *args: P.args, **kwargs: P.kwargs) -> R:
            self._load()
            return func(self, *args, **kwargs)

        return wrapper

    def _load(self):
        # maybe during deepcopy from field default
        if not self.__instance__:
            return self

        # or the relationship is already loaded
        if self.__loaded__:
            return self

        active_materia.get().load_association(self)

        # A: No provided, None
        # B: provided value is empty, {}
        if not self.__provided__:
            return self

        # Remove payloads whose key is already in __provided__
        self.__payloads__ = {
            k: v for k, v in self.__payloads__.items() if k not in self.__provided__
        }

        if len(self.__provided__) != super().__len__():
            # If the length of __provided__ is not equal to the length of self,
            # it means some items were not blessed into transmuter objects.
            super().clear()
            super().update(self.bless(self.__provided__))
        self.__loaded__ = True

        return self

    async def _aload(self):
        # maybe during deepcopy from field default
        if not self.__instance__:
            return self

        # or the relationship is already loaded
        if self.__loaded__:
            return self

        # A: No provided, None
        # B: provided value is empty, {}
        if not (provided := await active_materia.get().aload_association(self)):
            return self

        # Remove payloads whose key is already in provided
        self.__payloads__ = {
            k: v for k, v in self.__payloads__.items() if k not in provided
        }

        if len(provided) != super().__len__():
            # If the length of __provided__ is not equal to the length of self,
            # it means some items were not blessed into transmuter objects.
            super().clear()
            super().update(self.bless(provided))
        self.__loaded__ = True

        return self

    def __await__(self):
        return self._aload().__await__()

    @ensure_loaded
    def __getitem__(self, key: K) -> T:
        key = self.bless_key(key)
        return super().__getitem__(key)

    @ensure_loaded
    def __iter__(self):
        return super().__iter__()

    @ensure_loaded
    def __len__(self):
        return super().__len__()

    @ensure_loaded
    def __contains__(self, key: object) -> bool:
        key = self.bless_key(key)
        return super().__contains__(key)

    @ensure_loaded
    def __bool__(self):
        return super().__len__() > 0

    @ensure_loaded
    def __setitem__(self, key: K, value: T) -> None:
        key = self.bless_key(key)
        value = self.bless_value(value)
        if self.__provided__ is not None:
            self.__provided__[key] = value.__transmuter_provided__
        super().__setitem__(key, value)

    @ensure_loaded
    def __delitem__(self, key: K) -> None:
        key = self.bless_key(key)
        if self.__provided__ is not None:
            del self.__provided__[key]
        super().__delitem__(key)

    def __repr__(self):
        key_type = self.__args__[0]
        value_type = self.__args__[1]
        key_name = getattr(key_type, "__name__", repr(key_type))
        value_name = getattr(value_type, "__name__", repr(value_type))
        return f"RelationMap[{key_name}, {value_name}], instance={id(self.__instance__)}, size={super().__len__()}"

    @ensure_loaded
    def __str__(self):
        return super().__str__()

    @ensure_loaded
    def __eq__(self, other: object) -> bool:
        if isinstance(other, dict):
            return dict.__eq__(self, other)
        return False

    @ensure_loaded
    def __ne__(self, other: object) -> bool:
        if isinstance(other, dict):
            return dict.__ne__(self, other)
        return True

    @ensure_loaded
    def __or__(self, other: Mapping[K, T]) -> dict[K, T]:
        return dict.__or__(self.copy(), dict(other))

    @ensure_loaded
    def __ior__(self, other: Mapping[K, T]) -> Self:
        self.update(other)
        return self

    @ensure_loaded
    def __reversed__(self):
        return super().__reversed__()

    @ensure_loaded
    def get(self, key: K, default: T | None = None) -> T | None:
        return super().get(self.bless_key(key), default)

    @ensure_loaded
    def keys(self):
        return super().keys()

    @ensure_loaded
    def values(self):
        return super().values()

    @ensure_loaded
    def items(self):
        return super().items()

    @overload
    def pop(self, key: K) -> T: ...

    @overload
    def pop(self, key: K, default: T) -> T: ...

    @overload
    def pop(self, key: K, default: D) -> T | D: ...

    @ensure_loaded
    def pop(self, key: K, *args: Any) -> Any:
        """Remove specified key and return the corresponding value."""
        key = self.bless_key(key)
        item = super().pop(key, *args)
        if self.__provided__ is not None and key in self.__provided__:
            del self.__provided__[key]
        return item

    @ensure_loaded
    def popitem(self) -> tuple[K, T]:
        """Remove and return an arbitrary (key, value) pair. Raises KeyError if empty."""
        key, item = super().popitem()
        if self.__provided__ is not None:
            del self.__provided__[key]
        return key, item

    @overload
    def update(self, m: Mapping[K, T], /, **kwargs: T) -> None: ...
    @overload
    def update(self, m: Iterable[tuple[K, T]], /, **kwargs: T) -> None: ...
    @overload
    def update(self, **kwargs: T) -> None: ...

    @ensure_loaded
    def update(
        self,
        *args: Mapping[K, T] | Iterable[tuple[K, T]],
        **kwargs: T,
    ) -> None:
        """Update the dict with key-value pairs."""
        merged = {}
        if args:
            if isinstance(args[0], Mapping):
                merged.update(args[0])
            else:
                merged.update(dict(*args))
        if kwargs:
            merged.update(kwargs)

        if not merged:
            return

        blessed = self.bless(merged)
        if self.__provided__ is not None:
            self.__provided__.update(
                {key: value.__transmuter_provided__ for key, value in blessed.items()}
            )
        super().update(blessed)

    @overload
    def setdefault(self, key: K) -> T | None: ...
    @overload
    def setdefault(self, key: K, default: T) -> T: ...

    @ensure_loaded
    def setdefault(self, key: K, default: T | None = None) -> T | None:
        """If key is not in the dict, insert key with the default value."""
        key = self.bless_key(key)
        if key not in self:
            if default is not None:
                self[key] = default
        return super().get(key, default)

    @ensure_loaded
    def clear(self) -> None:
        """Remove all items."""
        if self.__provided__ is not None:
            self.__provided__.clear()
        super().clear()

    @ensure_loaded
    def copy(self) -> dict[K, T]:
        return super().copy()


# built-in types must be put at front to avoid pydantic convert it to built-in types
class TypedRelationMap(dict, Association[TD]):
    """A dict-based association whose keys and per-key value types are defined
    by a ``TypedDict``.

    Unlike :class:`RelationMap` which maps homogeneous ``dict[K, T]``, this
    class accepts a single ``TypedDict`` generic argument so that each key can
    have its own Transmuter type.  This is designed for polymorphic
    relationships where different keys correspond to different subclasses.

    Usage::

        class DocumentFiles(TypedDict):
            image: Image
            video: Video

        class Document(BaseTransmuter):
            files: TypedRelationMap[DocumentFiles] = TypedRelationMaps()

    All values in the TypedDict **must** be ``Transmuter`` subclasses.
    """

    # new items are held in __payloads__, loaded items are kept in the dict itself
    __payloads__: dict[str, Any]
    __typed_dict__: type  # the TypedDict class itself

    @classmethod
    def __get_pydantic_generic_schema__(
        cls,
        typed_dict_cls: type,
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        return handler.generate_schema(typed_dict_cls)

    @classmethod
    def __get_pydantic_serialize_schema__(
        cls,
        typed_dict_cls: type,
        handler: GetCoreSchemaHandler,
    ) -> core_schema.SerSchema | None:
        def serialize(association: TypedRelationMap, serializer) -> Any:
            fields_set = getattr(
                association.__instance__, "__pydantic_fields_set__", None
            )
            if fields_set is not None and association.field_name in fields_set:
                return serializer(association.copy())
            return serializer(dict.copy(association) | association.__payloads__)

        # Build a union schema from the TypedDict value types for serialization.
        # Using a plain dict schema avoids the TypedDict's totality constraint
        # so partial dicts (e.g. loaded from DB) serialize correctly.
        hints = get_type_hints(typed_dict_cls)
        value_types = list(dict.fromkeys(hints.values()))  # unique, order-preserved
        if len(value_types) == 1:
            values_schema = handler.generate_schema(value_types[0])
        else:
            values_schema = core_schema.union_schema(
                [handler.generate_schema(t) for t in value_types]
            )

        return core_schema.wrap_serializer_function_ser_schema(
            serialize,
            schema=core_schema.dict_schema(
                keys_schema=core_schema.str_schema(),
                values_schema=values_schema,
            ),
            when_used="always",
        )

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Type[TypedRelationMap],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        args = get_args(source_type)

        if not args or len(args) != 1:
            raise TypeError(
                f"Exactly one TypedDict generic type must be provided to {source_type}."
            )

        typed_dict_cls = args[0]

        if not is_typeddict(typed_dict_cls):
            raise TypeError(
                f"The generic argument to TypedRelationMap must be a TypedDict, "
                f"got {typed_dict_cls!r}."
            )

        def validate(
            value: Any,
            handler: core_schema.ValidatorFunctionWrapHandler,
            info: core_schema.ValidationInfo,
        ) -> TypedRelationMap:
            if value is DefferedAssociation:
                instance = cls()
            elif type(value) is cls:
                instance = value
                instance.__payloads__ = handler(instance.__payloads__)
            else:
                instance = cls(handler(value))

            instance.__args__ = (typed_dict_cls,)
            instance.__typed_dict__ = typed_dict_cls
            instance.field_name = info.field_name  # pyright: ignore[reportAttributeAccessIssue]

            return instance

        return core_schema.with_default_schema(
            core_schema.with_info_wrap_validator_function(
                validate,
                cls.__get_pydantic_generic_schema__(typed_dict_cls, handler),
            ),
            default_factory=cls,
            serialization=cls.__get_pydantic_serialize_schema__(
                typed_dict_cls, handler
            ),
        )

    @cached_property
    def __typed_hints__(self) -> dict[str, type]:
        """Per-key type hints resolved from the TypedDict."""
        return get_type_hints(self.__typed_dict__)

    @cached_property
    def __validator__(self) -> TypeAdapter:
        """Whole-dict TypedDict validator."""
        return get_cached_adapter(self.__typed_dict__)

    def bless(self, value: Mapping[str, Any]) -> dict[str, Any]:
        """Validate and coerce an entire mapping against the TypedDict schema."""
        return self.__validator__.validate_python(value)

    def bless_value(self, key: str, value: Any) -> Any:
        """Validate a single value against the type declared for *key*."""
        expected_type = self.__typed_hints__.get(key)
        if expected_type is None:
            raise KeyError(
                f"Key {key!r} is not defined in {self.__typed_dict__.__name__}."
            )
        return get_cached_adapter(expected_type).validate_python(value)

    def __init__(self, payloads: Mapping[str, Any] | None = None):
        super().__init__()
        self.__instance__ = None
        self.__loaded__ = False
        self.__payloads__ = dict(payloads) if payloads else {}

    @property
    def __provided__(self) -> dict | None:
        if not self.__instance_provider__:
            return None
        return getattr(self.__instance_provider__, self.used_name)

    def prepare(self, instance: Transmuter, field_name: str):
        if self.__instance__ is not None:
            return

        self.field_name = field_name
        self.field_info = type(instance).__pydantic_fields__[field_name]

        self.__instance__ = instance

        annotation = self.field_info.annotation
        if isinstance(annotation, ForwardRef):
            resolved_hints = get_type_hints(type(instance))
            actual_type = resolved_hints[field_name]
            args = get_args(actual_type)
        else:
            args = get_args(annotation)

        self.__typed_dict__ = args[0]
        self.__args__ = (args[0],)

        if self.__payloads__:
            self._load()
            self.update(self.__payloads__)
            self.__payloads__.clear()

    @staticmethod
    def ensure_loaded(
        func: Callable[Concatenate[TypedRelationMap, P], R],
    ) -> Callable[Concatenate[TypedRelationMap, P], R]:
        @wraps(func)
        def wrapper(self: TypedRelationMap, *args: P.args, **kwargs: P.kwargs) -> R:
            self._load()
            return func(self, *args, **kwargs)

        return wrapper

    def _load(self):
        if not self.__instance__:
            return self

        if self.__loaded__:
            return self

        active_materia.get().load_association(self)

        if not self.__provided__:
            return self

        # Remove payloads whose key is already in __provided__
        self.__payloads__ = {
            k: v for k, v in self.__payloads__.items() if k not in self.__provided__
        }

        if len(self.__provided__) != super().__len__():
            super().clear()
            # Validate per-key: DB state may be partial (not all TypedDict keys present)
            super().update(
                {k: self.bless_value(k, v) for k, v in self.__provided__.items()}
            )
        self.__loaded__ = True

        return self

    async def _aload(self):
        if not self.__instance__:
            return self

        if self.__loaded__:
            return self

        if not (provided := await active_materia.get().aload_association(self)):
            return self

        self.__payloads__ = {
            k: v for k, v in self.__payloads__.items() if k not in provided
        }

        if len(provided) != super().__len__():
            super().clear()
            # Validate per-key: DB state may be partial (not all TypedDict keys present)
            super().update({k: self.bless_value(k, v) for k, v in provided.items()})
        self.__loaded__ = True

        return self

    def __await__(self):
        return self._aload().__await__()

    @ensure_loaded
    def __getitem__(self, key: str) -> Any:
        return super().__getitem__(key)

    @ensure_loaded
    def __iter__(self):
        return super().__iter__()

    @ensure_loaded
    def __len__(self):
        return super().__len__()

    @ensure_loaded
    def __contains__(self, key: object) -> bool:
        return super().__contains__(key)

    @ensure_loaded
    def __bool__(self):
        return super().__len__() > 0

    @ensure_loaded
    def __setitem__(self, key: str, value: Any) -> None:
        value = self.bless_value(key, value)
        if self.__provided__ is not None:
            self.__provided__[key] = value.__transmuter_provided__
        super().__setitem__(key, value)

    @ensure_loaded
    def __delitem__(self, key: str) -> None:
        if self.__provided__ is not None:
            del self.__provided__[key]
        super().__delitem__(key)

    def __repr__(self):
        td_name = getattr(self.__typed_dict__, "__name__", repr(self.__typed_dict__))
        return f"TypedRelationMap[{td_name}], instance={id(self.__instance__)}, size={super().__len__()}"

    @ensure_loaded
    def __str__(self):
        return dict.__repr__(self)

    @ensure_loaded
    def __eq__(self, other: object) -> bool:
        if isinstance(other, dict):
            return dict.__eq__(self, other)
        return False

    @ensure_loaded
    def __ne__(self, other: object) -> bool:
        if isinstance(other, dict):
            return dict.__ne__(self, other)
        return True

    @ensure_loaded
    def __or__(self, other: Mapping[str, Any]) -> dict[str, Any]:
        return dict.__or__(self.copy(), dict(other))

    @ensure_loaded
    def __ior__(self, other: Mapping[str, Any]) -> Self:
        self.update(other)
        return self

    @ensure_loaded
    def __reversed__(self):
        return super().__reversed__()

    @ensure_loaded
    def get(self, key: str, default: Any = None) -> Any:
        return super().get(key, default)

    @ensure_loaded
    def keys(self):
        return super().keys()

    @ensure_loaded
    def values(self):
        return super().values()

    @ensure_loaded
    def items(self):
        return super().items()

    @ensure_loaded
    def pop(self, key: str, *args: Any) -> Any:
        """Remove specified key and return the corresponding value."""
        item = super().pop(key, *args)
        if self.__provided__ is not None and key in self.__provided__:
            del self.__provided__[key]
        return item

    @ensure_loaded
    def popitem(self) -> tuple[str, Any]:
        """Remove and return an arbitrary (key, value) pair."""
        key, item = super().popitem()
        if self.__provided__ is not None:
            del self.__provided__[key]
        return key, item

    @ensure_loaded
    def update(
        self,
        *args: Mapping[str, Any] | Iterable[tuple[str, Any]],
        **kwargs: Any,
    ) -> None:
        """Update the dict with key-value pairs, validating per-key."""
        merged: dict[str, Any] = {}
        if args:
            if isinstance(args[0], Mapping):
                merged.update(args[0])
            else:
                merged.update(dict(*args))
        if kwargs:
            merged.update(kwargs)

        if not merged:
            return

        # Validate per-key to allow partial updates (not all TypedDict keys required)
        blessed: dict[str, Any] = {}
        for key, value in merged.items():
            blessed[key] = self.bless_value(key, value)
        if self.__provided__ is not None:
            self.__provided__.update(
                {key: value.__transmuter_provided__ for key, value in blessed.items()}
            )
        super().update(blessed)

    @ensure_loaded
    def setdefault(self, key: str, default: Any = None) -> Any:
        """If key is not in the dict, insert key with the default value."""
        if key not in self:
            if default is not None:
                self[key] = default
        return super().get(key, default)

    @ensure_loaded
    def clear(self) -> None:
        """Remove all items."""
        if self.__provided__ is not None:
            self.__provided__.clear()
        super().clear()

    @ensure_loaded
    def copy(self) -> dict[str, Any]:
        return super().copy()


if TYPE_CHECKING:
    from typing_extensions import TypeAliasType as TypeAliasType

    # Make TypedRelationMap[TD] resolve to TD for the type checker.
    # This lets Pyright/mypy apply TypedDict per-key type inference so that
    # e.g. ``gallery.media["image"]`` returns ``ImageMedia`` instead of ``Any``.
    TypedRelationMap = TypeAliasType(  # type: ignore[assignment]
        "TypedRelationMap", TD, type_params=(TD,)
    )

Relationship = partial(Field, default_factory=Relation, frozen=True)
RelationMaps = partial(Field, default_factory=RelationMap, frozen=True)
TypedRelationMaps = partial(Field, default_factory=TypedRelationMap, frozen=True)


@overload
def Relationships(*, unique: Literal[True], **kwargs: Any) -> Any: ...
@overload
def Relationships(*, unique: Literal[False] = ..., **kwargs: Any) -> Any: ...
@overload
def Relationships(**kwargs: Any) -> Any: ...
def Relationships(*, unique: bool = False, **kwargs: Any) -> Any:
    """Create a relationship field for a collection of related transmuters.

    Args:
        unique: If True, use a RelationSet (set semantics, no duplicates).
                If False (default), use a RelationCollection (list semantics).
        **kwargs: Additional keyword arguments passed to pydantic's Field().

    Returns:
        A pydantic Field configured with the appropriate default_factory.
    """
    factory = RelationSet if unique else RelationCollection
    return Field(default_factory=factory, frozen=True, **kwargs)



"""Tests for dataclass transmuters with NoOpMateria.

Covers:
- Basic creation (direct and via TypeAdapter)
- Field defaults and validation
- Identity fields
- Association fields (Relation, RelationCollection)
- validate_assignment support
- Revalidation
- validation_context
- model_dump / serialization
- Create / Update partial models
- __setattr__ sync behavior
- Same-type passthrough
- Metaclass introspection
"""

from __future__ import annotations

import pytest
from pydantic import TypeAdapter, ValidationError

from arcanus.association import Relation, RelationCollection
from arcanus.base import (
    Transmuter,
    TransmuterMetaclass,
    validation_context,
)
from tests.transmuters import (
    DCAuthor,
    DCBook,
    IdentifiedUser,
    PydanticItem,
    SimpleTag,
    SimpleUser,
    ValidatedUser,
)


class TestBasicCreation:
    """Test basic dataclass transmuter creation."""

    def test_direct_creation(self):
        user = SimpleUser(name="Alice", age=30)
        assert user.name == "Alice"
        assert user.age == 30

    def test_direct_creation_with_defaults(self):
        user = SimpleUser(name="Bob")
        assert user.name == "Bob"
        assert user.age == 25  # default

    def test_type_adapter_from_dict(self):
        ta = TypeAdapter(SimpleUser)
        user = ta.validate_python({"name": "Carol", "age": 40})
        assert user.name == "Carol"
        assert user.age == 40

    def test_type_adapter_partial_dict(self):
        ta = TypeAdapter(SimpleUser)
        user = ta.validate_python({"name": "Dan"})
        assert user.name == "Dan"
        assert user.age == 25

    def test_type_adapter_same_type_passthrough(self):
        ta = TypeAdapter(SimpleUser)
        user = SimpleUser(name="Eve")
        result = ta.validate_python(user)
        assert result is user

    def test_validation_error_on_invalid_data(self):
        ta = TypeAdapter(SimpleUser)
        with pytest.raises(ValidationError):
            ta.validate_python({"name": 123})  # name should be str

    def test_validation_error_on_missing_required(self):
        ta = TypeAdapter(SimpleUser)
        with pytest.raises(ValidationError):
            ta.validate_python({})


class TestTransmuterAttributes:
    """Test transmuter instance attributes (__transmuter_provided__, etc.)."""

    def test_provided_is_none_in_noop(self):
        user = SimpleUser(name="Alice")
        assert user.__transmuter_provided__ is None

    def test_revalidating_is_false(self):
        user = SimpleUser(name="Alice")
        assert user.__transmuter_revalidating__ is False


class TestMetaclassIntrospection:
    """Test TransmuterMetaclass properties on dataclass transmuters."""

    def test_is_dataclass(self):
        assert SimpleUser.__transmuter_is_dataclass__ is True

    def test_transmuter_complete(self):
        assert SimpleUser.__transmuter_complete__ is True

    def test_pydantic_fields(self):
        assert set(SimpleUser.__pydantic_fields__.keys()) == {"name", "age"}

    def test_model_associations_empty(self):
        assert SimpleUser.model_associations == {}

    def test_model_identities_empty(self):
        assert SimpleUser.model_identities == {}

    def test_model_associations_with_relations(self):
        assert "author" in DCBook.model_associations
        assert "publisher" in DCBook.model_associations
        assert "tags" in DCBook.model_associations

    def test_model_identities_with_identity(self):
        assert "id" in IdentifiedUser.model_identities

    def test_create_model(self):
        Create = SimpleUser.Create
        assert "name" in Create.model_fields
        assert "age" in Create.model_fields

    def test_create_model_excludes_identities(self):
        Create = IdentifiedUser.Create
        assert "id" not in Create.model_fields
        assert "name" in Create.model_fields
        assert "email" in Create.model_fields

    def test_update_model(self):
        Update = SimpleUser.Update
        assert "name" in Update.model_fields
        assert "age" in Update.model_fields


class TestIdentityFields:
    """Test identity field tracking."""

    def test_identity_detection(self):
        assert "id" in IdentifiedUser.model_identities

    def test_identity_field_create(self):
        user = IdentifiedUser(name="Alice", email="alice@example.com")
        assert user.name == "Alice"
        assert user.id is None  # default

    def test_identity_field_set(self):
        user = IdentifiedUser(name="Alice", id=42)
        assert user.id == 42


class TestValidateAssignment:
    """Test validate_assignment support."""

    def test_setattr_updates_value(self):
        user = ValidatedUser(name="Alice", age=30)
        user.name = "Bob"
        assert user.name == "Bob"

    def test_setattr_validates(self):
        user = ValidatedUser(name="Alice", age=30)
        with pytest.raises(ValidationError):
            user.age = "not a number"  # type: ignore


class TestRevalidation:
    """Test revalidate behavior."""

    def test_revalidate_returns_self(self):
        user = SimpleUser(name="Alice")
        result = user.revalidate()
        assert result is user

    def test_revalidate_no_error_without_provider(self):
        """With NoOp, revalidate should be a no-op."""
        user = SimpleUser(name="Alice")
        user.revalidate()  # should not raise


class TestValidationContext:
    """Test validation_context integration."""

    def test_validation_context_works(self):
        ta = TypeAdapter(SimpleUser)
        with validation_context():
            user = ta.validate_python({"name": "Alice"})
            assert user.name == "Alice"


class TestSerialization:
    """Test TypeAdapter dump_python for dataclass transmuters."""

    def test_dump_simple(self):
        user = SimpleUser(name="Alice", age=30)
        ta = TypeAdapter(SimpleUser)
        dumped = ta.dump_python(user)
        assert dumped == {"name": "Alice", "age": 30}

    def test_dump_with_associations(self):
        ta_book = TypeAdapter(DCBook)
        book = ta_book.validate_python(
            {
                "title": "Test Book",
                "year": 2024,
                "author": {"name": "Alice", "field": "Physics"},
                "tags": [{"label": "sci"}, {"label": "tech"}],
            }
        )

        dumped = ta_book.dump_python(book)
        assert dumped["title"] == "Test Book"
        assert dumped["year"] == 2024
        assert dumped["author"]["name"] == "Alice"
        assert len(dumped["tags"]) == 2


class TestRelationBasics:
    """Test Relation fields on dataclass transmuters."""

    def test_relation_initialization_empty(self):
        book = DCBook(title="Test")
        assert isinstance(book.author, Relation)
        assert book.author.value is None

    def test_relation_initialization_with_value_via_validation(self):
        ta = TypeAdapter(DCBook)
        book = ta.validate_python(
            {
                "title": "Test",
                "author": {"name": "Alice", "field": "Physics"},
            }
        )
        assert isinstance(book.author, Relation)
        assert book.author.value is not None
        assert book.author.value.name == "Alice"

    def test_relation_set_value(self):
        book = DCBook(title="Test")
        author = DCAuthor(name="Bob", field="Biology")
        book.author.value = author
        assert book.author.value.name == "Bob"

    def test_relation_isinstance(self):
        book = DCBook(title="Test")
        assert isinstance(book.author, Relation)


class TestRelationCollectionBasics:
    """Test RelationCollection fields on dataclass transmuters."""

    def test_collection_initialization_empty(self):
        book = DCBook(title="Test")
        assert isinstance(book.tags, RelationCollection)
        assert len(book.tags) == 0

    def test_collection_initialization_with_values_via_validation(self):
        ta = TypeAdapter(DCBook)
        book = ta.validate_python(
            {
                "title": "Test",
                "tags": [{"label": "sci"}, {"label": "tech"}],
            }
        )
        assert isinstance(book.tags, RelationCollection)
        assert len(book.tags) == 2
        assert book.tags[0].label == "sci"

    def test_collection_append(self):
        book = DCBook(title="Test")
        tag = SimpleTag(label="new")
        book.tags.append(tag)
        assert len(book.tags) == 1
        assert book.tags[0].label == "new"

    def test_collection_iteration(self):
        ta = TypeAdapter(DCBook)
        book = ta.validate_python(
            {
                "title": "Test",
                "tags": [{"label": "a"}, {"label": "b"}, {"label": "c"}],
            }
        )
        labels = [t.label for t in book.tags]
        assert labels == ["a", "b", "c"]


class TestInstanceIsolation:
    """Test that transmuter instances are properly isolated."""

    def test_separate_instances_independent(self):
        u1 = SimpleUser(name="Alice")
        u2 = SimpleUser(name="Bob")
        assert u1.name != u2.name

    def test_association_isolation(self):
        """Each instance should have its own association."""
        b1 = DCBook(title="Book 1")
        b2 = DCBook(title="Book 2")

        tag = SimpleTag(label="shared")
        b1.tags.append(tag)

        assert len(b1.tags) == 1
        assert len(b2.tags) == 0


class TestTransmuterInstanceMixin:
    """Verify Transmuter is in the hierarchy and provides methods."""

    def test_plain_dc_has_mixin(self):
        assert issubclass(SimpleUser, Transmuter)

    def test_mixin_provides_revalidate(self):
        u = SimpleUser(name="Alice")
        assert hasattr(u, "revalidate")
        assert u.revalidate() is u


class TestPydanticDataclassBlessing:
    """Test pydantic @dataclass blessed via NoOpMateria().bless()."""

    def test_create_from_pydantic(self):
        item = PydanticItem(name="test")
        assert item.name == "test"
        assert item.score == 0.0

    def test_is_transmuter_metaclass(self):
        assert isinstance(PydanticItem, TransmuterMetaclass)

    def test_type_adapter_works(self):
        adapter = TypeAdapter(PydanticItem)
        item = adapter.validate_python({"name": "test", "score": 9.5})
        assert isinstance(item, PydanticItem)
        assert item.score == 9.5

    def test_has_mixin(self):
        assert issubclass(PydanticItem, Transmuter)

    def test_revalidate(self):
        item = PydanticItem(name="test")
        assert item.revalidate() is item

    def test_transmuter_attributes(self):
        assert hasattr(PydanticItem, "__pydantic_fields__")
        assert hasattr(PydanticItem, "model_associations")
        assert "name" in PydanticItem.__pydantic_fields__
        assert "score" in PydanticItem.__pydantic_fields__



"""Test RelationMap association fields with NoOpMateria."""

from __future__ import annotations

from typing import Annotated, Optional

import pytest
from pydantic import ConfigDict, Field, ValidationError

from arcanus.association import RelationMap, RelationMaps
from arcanus.base import BaseTransmuter, Identity
from tests.transmuters import Catalog, LabeledCatalog, Tag


class TestRelationMapBasics:
    """Test basic RelationMap field behaviour."""

    def test_init_empty(self):
        catalog = Catalog(id=1, title="Test")
        assert isinstance(catalog.tags, RelationMap)
        assert len(catalog.tags) == 0

    def test_init_with_values(self):
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="rust")
        catalog = Catalog(id=1, title="Test", tags=RelationMap({"py": t1, "rs": t2}))
        assert isinstance(catalog.tags, RelationMap)
        assert len(catalog.tags) == 2

    def test_setitem(self):
        catalog = Catalog(id=1, title="Test")
        t = Tag(id=1, name="python")
        catalog.tags["py"] = t
        assert len(catalog.tags) == 1
        assert catalog.tags["py"] is t

    def test_getitem(self):
        catalog = Catalog(id=1, title="Test")
        t = Tag(id=1, name="python")
        catalog.tags["py"] = t
        assert catalog.tags["py"].name == "python"

    def test_getitem_missing_raises(self):
        catalog = Catalog(id=1, title="Test")
        with pytest.raises(KeyError):
            catalog.tags["missing"]

    def test_delitem(self):
        catalog = Catalog(id=1, title="Test")
        t = Tag(id=1, name="python")
        catalog.tags["py"] = t
        del catalog.tags["py"]
        assert len(catalog.tags) == 0

    def test_delitem_missing_raises(self):
        catalog = Catalog(id=1, title="Test")
        with pytest.raises(KeyError):
            del catalog.tags["missing"]

    def test_get_present(self):
        catalog = Catalog(id=1, title="Test")
        t = Tag(id=1, name="python")
        catalog.tags["py"] = t
        assert catalog.tags.get("py") is t

    def test_get_absent_returns_default(self):
        catalog = Catalog(id=1, title="Test")
        assert catalog.tags.get("missing") is None
        sentinel = Tag(id=99, name="default")
        assert catalog.tags.get("missing", sentinel) is sentinel

    def test_pop_present(self):
        catalog = Catalog(id=1, title="Test")
        t = Tag(id=1, name="python")
        catalog.tags["py"] = t
        popped = catalog.tags.pop("py")
        assert popped.name == "python"
        assert len(catalog.tags) == 0

    def test_pop_absent_raises(self):
        catalog = Catalog(id=1, title="Test")
        with pytest.raises(KeyError):
            catalog.tags.pop("missing")

    def test_pop_absent_with_default(self):
        catalog = Catalog(id=1, title="Test")
        sentinel = Tag(id=99, name="default")
        result = catalog.tags.pop("missing", sentinel)
        assert result is sentinel

    def test_popitem(self):
        catalog = Catalog(id=1, title="Test")
        catalog.tags["py"] = Tag(id=1, name="python")
        key, val = catalog.tags.popitem()
        assert key == "py"
        assert val.name == "python"
        assert len(catalog.tags) == 0

    def test_popitem_empty_raises(self):
        catalog = Catalog(id=1, title="Test")
        with pytest.raises(KeyError):
            catalog.tags.popitem()

    def test_clear(self):
        catalog = Catalog(id=1, title="Test")
        for i in range(3):
            catalog.tags[f"t{i}"] = Tag(id=i, name=f"tag{i}")
        assert len(catalog.tags) == 3
        catalog.tags.clear()
        assert len(catalog.tags) == 0

    def test_setdefault_absent(self):
        catalog = Catalog(id=1, title="Test")
        t = Tag(id=1, name="python")
        result = catalog.tags.setdefault("py", t)
        assert result is t
        assert catalog.tags["py"] is t

    def test_setdefault_present(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="rust")
        catalog.tags["py"] = t1
        result = catalog.tags.setdefault("py", t2)
        assert result is t1


class TestRelationMapUpdate:
    """Test update operations."""

    def test_update_with_dict(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="rust")
        catalog.tags.update({"py": t1, "rs": t2})
        assert len(catalog.tags) == 2
        assert catalog.tags["py"].name == "python"
        assert catalog.tags["rs"].name == "rust"

    def test_update_with_kwargs(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        catalog.tags.update(py=t1)
        assert catalog.tags["py"].name == "python"

    def test_update_overwrites(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="python3")
        catalog.tags["py"] = t1
        catalog.tags.update({"py": t2})
        assert catalog.tags["py"] is t2


class TestRelationMapKeysValuesItems:
    """Test keys(), values(), items() views."""

    def test_keys(self):
        catalog = Catalog(id=1, title="Test")
        catalog.tags["py"] = Tag(id=1, name="python")
        catalog.tags["rs"] = Tag(id=2, name="rust")
        assert set(catalog.tags.keys()) == {"py", "rs"}

    def test_values(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="rust")
        catalog.tags["py"] = t1
        catalog.tags["rs"] = t2
        vals = list(catalog.tags.values())
        assert len(vals) == 2
        assert t1 in vals
        assert t2 in vals

    def test_items(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        catalog.tags["py"] = t1
        items = list(catalog.tags.items())
        assert items == [("py", t1)]


class TestRelationMapComparisons:
    """Test equality and comparison operations."""

    def test_eq_same_content(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        catalog.tags["py"] = t1
        assert catalog.tags == {"py": t1}

    def test_ne_different_content(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="rust")
        catalog.tags["py"] = t1
        assert catalog.tags != {"rs": t2}

    def test_eq_empty(self):
        catalog = Catalog(id=1, title="Test")
        assert catalog.tags == {}

    def test_ne_non_dict(self):
        catalog = Catalog(id=1, title="Test")
        assert catalog.tags != [1, 2, 3]
        assert catalog.tags != "hello"

    def test_eq_two_relation_maps(self):
        c1 = Catalog(id=1, title="A")
        c2 = Catalog(id=2, title="B")
        t = Tag(id=1, name="python")
        c1.tags["py"] = t
        c2.tags["py"] = t
        assert c1.tags == c2.tags


class TestRelationMapIteration:
    """Test iteration and membership operations."""

    def test_iter_yields_keys(self):
        catalog = Catalog(id=1, title="Test")
        catalog.tags["py"] = Tag(id=1, name="python")
        catalog.tags["rs"] = Tag(id=2, name="rust")

        collected = list(catalog.tags)
        assert set(collected) == {"py", "rs"}

    def test_contains_key(self):
        catalog = Catalog(id=1, title="Test")
        catalog.tags["py"] = Tag(id=1, name="python")
        assert "py" in catalog.tags
        assert "rs" not in catalog.tags

    def test_len(self):
        catalog = Catalog(id=1, title="Test")
        assert len(catalog.tags) == 0
        catalog.tags["py"] = Tag(id=1, name="python")
        assert len(catalog.tags) == 1
        catalog.tags["rs"] = Tag(id=2, name="rust")
        assert len(catalog.tags) == 2

    def test_bool_empty(self):
        catalog = Catalog(id=1, title="Test")
        assert not bool(catalog.tags)

    def test_bool_nonempty(self):
        catalog = Catalog(id=1, title="Test")
        catalog.tags["py"] = Tag(id=1, name="python")
        assert bool(catalog.tags)

    def test_reversed(self):
        catalog = Catalog(id=1, title="Test")
        catalog.tags["a"] = Tag(id=1, name="alpha")
        catalog.tags["b"] = Tag(id=2, name="beta")
        catalog.tags["c"] = Tag(id=3, name="gamma")
        keys = list(reversed(catalog.tags))
        assert keys == ["c", "b", "a"]


class TestRelationMapInstanceIsolation:
    """Test that RelationMap instances are not shared between model instances."""

    def test_separate_instances_independent(self):
        c1 = Catalog(id=1, title="Catalog 1")
        c2 = Catalog(id=2, title="Catalog 2")

        t1 = Tag(id=1, name="tag1")
        t2 = Tag(id=2, name="tag2")

        c1.tags["a"] = t1
        c2.tags["b"] = t2

        assert len(c1.tags) == 1
        assert len(c2.tags) == 1
        assert "a" in c1.tags
        assert "b" in c2.tags
        assert "a" not in c2.tags
        assert "b" not in c1.tags

    def test_clear_one_does_not_affect_other(self):
        c1 = Catalog(id=1, title="Catalog 1")
        c2 = Catalog(id=2, title="Catalog 2")

        t = Tag(id=1, name="shared")
        c1.tags["x"] = t
        c2.tags["x"] = t

        c1.tags.clear()
        assert len(c1.tags) == 0
        assert len(c2.tags) == 1


class TestRelationMapSerialization:
    """Test serialization / deserialization with model_dump / model_validate."""

    def test_model_dump_empty(self):
        catalog = Catalog(id=1, title="Test")
        data = catalog.model_dump()
        assert data["tags"] == {}

    def test_model_dump_with_items(self):
        catalog = Catalog(
            id=1,
            title="Test",
            tags=RelationMap(
                {
                    "py": Tag(id=1, name="python"),
                    "rs": Tag(id=2, name="rust"),
                }
            ),
        )

        data = catalog.model_dump()
        assert isinstance(data["tags"], dict)
        assert set(data["tags"].keys()) == {"py", "rs"}
        assert data["tags"]["py"]["name"] == "python"
        assert data["tags"]["rs"]["name"] == "rust"

    def test_model_validate_with_dict_input(self):
        catalog_data = {
            "id": 1,
            "title": "Test",
            "tags": {
                "py": {"id": 1, "name": "python"},
                "rs": {"id": 2, "name": "rust"},
            },
        }
        catalog = Catalog.model_validate(catalog_data)
        assert isinstance(catalog.tags, RelationMap)
        assert len(catalog.tags) == 2
        assert catalog.tags["py"].name == "python"
        assert catalog.tags["rs"].name == "rust"

    def test_isinstance_check(self):
        catalog = Catalog(id=1, title="Test")
        assert isinstance(catalog.tags, RelationMap)


class TestRelationMapMergeOperators:
    """Test | and |= operators."""

    def test_or_operator(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="rust")
        catalog.tags["py"] = t1
        result = catalog.tags | {"rs": t2}
        assert isinstance(result, dict)
        assert len(result) == 2
        assert result["py"] is t1
        assert result["rs"] is t2

    def test_ior_operator(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="rust")
        catalog.tags["py"] = t1
        tags = catalog.tags
        tags |= {"rs": t2}
        assert len(catalog.tags) == 2
        assert "rs" in catalog.tags

    def test_or_does_not_mutate_original(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="rust")
        catalog.tags["py"] = t1
        _ = catalog.tags | {"rs": t2}
        assert len(catalog.tags) == 1


class TestRelationMapCopy:
    """Test copy operations."""

    def test_copy_returns_plain_dict(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="rust")
        catalog.tags["py"] = t1
        catalog.tags["rs"] = t2

        result = catalog.tags.copy()
        assert isinstance(result, dict)
        assert not isinstance(result, RelationMap)
        assert len(result) == 2

    def test_repr(self):
        catalog = Catalog(id=1, title="Test")
        r = repr(catalog.tags)
        assert "RelationMap" in r
        assert "Tag" in r


class TestRelationMapEdgeCases:
    """Test edge cases."""

    def test_empty_map_operations(self):
        catalog = Catalog(id=1, title="Test")

        assert len(catalog.tags) == 0
        assert list(catalog.tags) == []
        assert catalog.tags.copy() == {}

    def test_multiple_relation_maps_on_same_model(self):
        """Each RelationMap field should be independent."""

        class MultiCatalog(BaseTransmuter):
            model_config = ConfigDict(from_attributes=True)

            id: Annotated[Optional[int], Identity] = Field(default=None, frozen=True)
            title: str
            tags: RelationMap[str, Tag] = RelationMaps()
            labels: RelationMap[str, Tag] = RelationMaps()

        cat = MultiCatalog(id=1, title="Test")
        t1 = Tag(id=1, name="a")
        t2 = Tag(id=2, name="b")

        cat.tags["x"] = t1
        cat.labels["y"] = t2

        assert len(cat.tags) == 1
        assert len(cat.labels) == 1
        assert "x" in cat.tags
        assert "y" in cat.labels
        assert "x" not in cat.labels
        assert "y" not in cat.tags

    def test_overwrite_existing_key(self):
        catalog = Catalog(id=1, title="Test")
        t1 = Tag(id=1, name="python")
        t2 = Tag(id=2, name="python3")
        catalog.tags["py"] = t1
        catalog.tags["py"] = t2
        assert len(catalog.tags) == 1
        assert catalog.tags["py"] is t2

    def test_str(self):
        catalog = Catalog(id=1, title="Test")
        s = str(catalog.tags)
        assert isinstance(s, str)


class TestRelationMapLiteralKeys:
    """Test RelationMap with Literal key type to verify key blessing."""

    def test_valid_literal_key_setitem(self):
        cat = LabeledCatalog(id=1, title="Test")
        t = Tag(id=1, name="py")
        cat.tags["python"] = t
        assert cat.tags["python"] is t

    def test_invalid_literal_key_setitem_raises(self):
        cat = LabeledCatalog(id=1, title="Test")
        t = Tag(id=1, name="x")
        with pytest.raises(ValidationError):
            cat.tags["java"] = t  # type: ignore[index]

    def test_valid_literal_key_getitem(self):
        cat = LabeledCatalog(id=1, title="Test")
        cat.tags["rust"] = Tag(id=1, name="rs")
        assert cat.tags["rust"].name == "rs"

    def test_invalid_literal_key_getitem_raises(self):
        cat = LabeledCatalog(id=1, title="Test")
        with pytest.raises(ValidationError):
            cat.tags["java"]  # type: ignore[index]

    def test_valid_literal_key_contains(self):
        cat = LabeledCatalog(id=1, title="Test")
        cat.tags["go"] = Tag(id=1, name="g")
        assert "go" in cat.tags

    def test_invalid_literal_key_contains_raises(self):
        cat = LabeledCatalog(id=1, title="Test")
        with pytest.raises(ValidationError):
            "java" in cat.tags  # type: ignore[operator]

    def test_valid_literal_key_delitem(self):
        cat = LabeledCatalog(id=1, title="Test")
        cat.tags["python"] = Tag(id=1, name="py")
        del cat.tags["python"]
        assert len(cat.tags) == 0

    def test_invalid_literal_key_delitem_raises(self):
        cat = LabeledCatalog(id=1, title="Test")
        with pytest.raises(ValidationError):
            del cat.tags["java"]  # type: ignore[arg-type]

    def test_valid_literal_key_get(self):
        cat = LabeledCatalog(id=1, title="Test")
        cat.tags["rust"] = Tag(id=1, name="rs")
        assert cat.tags["rust"].name == "rs"
        assert cat.tags.get("python") is None

    def test_invalid_literal_key_get_raises(self):
        cat = LabeledCatalog(id=1, title="Test")
        with pytest.raises(ValidationError):
            cat.tags.get("java")  # type: ignore[arg-type]

    def test_valid_literal_key_pop(self):
        cat = LabeledCatalog(id=1, title="Test")
        cat.tags["go"] = Tag(id=1, name="g")
        popped = cat.tags.pop("go")
        assert popped.name == "g"
        assert len(cat.tags) == 0

    def test_invalid_literal_key_pop_raises(self):
        cat = LabeledCatalog(id=1, title="Test")
        with pytest.raises(ValidationError):
            cat.tags.pop("java")  # type: ignore[arg-type]

    def test_valid_literal_key_setdefault(self):
        cat = LabeledCatalog(id=1, title="Test")
        t = Tag(id=1, name="py")
        result = cat.tags.setdefault("python", t)
        assert result is t
        assert cat.tags["python"] is t

    def test_invalid_literal_key_setdefault_raises(self):
        cat = LabeledCatalog(id=1, title="Test")
        t = Tag(id=1, name="x")
        with pytest.raises(ValidationError):
            cat.tags.setdefault("java", t)  # type: ignore[arg-type]

    def test_all_valid_keys(self):
        cat = LabeledCatalog(id=1, title="Test")
        cat.tags["python"] = Tag(id=1, name="py")
        cat.tags["rust"] = Tag(id=2, name="rs")
        cat.tags["go"] = Tag(id=3, name="g")
        assert len(cat.tags) == 3
        assert set(cat.tags.keys()) == {"python", "rust", "go"}

    def test_update_with_valid_keys(self):
        cat = LabeledCatalog(id=1, title="Test")
        cat.tags.update({"python": Tag(id=1, name="py"), "rust": Tag(id=2, name="rs")})
        assert len(cat.tags) == 2

    def test_update_with_invalid_key_raises(self):
        cat = LabeledCatalog(id=1, title="Test")
        with pytest.raises(ValidationError):
            cat.tags.update({"java": Tag(id=1, name="j")})  # type: ignore[dict-item]

    def test_init_with_valid_keys(self):
        t = Tag(id=1, name="py")
        cat = LabeledCatalog(
            id=1,
            title="Test",
            tags=RelationMap({"python": t}),
        )
        assert len(cat.tags) == 1
        assert cat.tags["python"] is t

    def test_init_with_invalid_key_raises(self):
        t = Tag(id=1, name="x")
        with pytest.raises(ValidationError):
            LabeledCatalog(
                id=1,
                title="Test",
                tags={"java": t},  # type: ignore[dict-item]
            )

    def test_model_validate_with_valid_keys(self):
        data = {
            "id": 1,
            "title": "Test",
            "tags": {"python": {"id": 1, "name": "py"}},
        }
        cat = LabeledCatalog.model_validate(data)
        assert len(cat.tags) == 1
        assert cat.tags["python"].name == "py"

    def test_model_validate_with_invalid_key_raises(self):
        data = {
            "id": 1,
            "title": "Test",
            "tags": {"java": {"id": 1, "name": "j"}},
        }
        with pytest.raises(ValidationError):
            LabeledCatalog.model_validate(data)



"""Test TypedRelationMap association with SQLAlchemy Materia.

Tests:
- CRUD operations with polymorphic dict-based relationships
  (Gallery <-> MediaAttachment via attribute_keyed_dict + single-table inheritance)
- Lazy and eager loading of typed dict relationships
- Serialization after ORM load
- Dict mutation operations persisted through SQLAlchemy
"""

from __future__ import annotations

from sqlalchemy import Engine, select
from sqlalchemy.orm import selectinload

from arcanus.association import TypedRelationMap
from arcanus.materia.sqlalchemy import Session
from tests.transmuters import Gallery, ImageMedia, VideoMedia


class TestTypedRelationMapCRUD:
    """Test basic CRUD operations with typed dict-based relationships."""

    def test_create_gallery_with_media(self, engine: Engine):
        """Test creating a Gallery with polymorphic media through TypedRelationMap."""
        with Session(engine) as session:
            gallery = Gallery(name="My Gallery")
            image = ImageMedia(slot="image", name="photo.jpg", width=1920, height=1080)
            video = VideoMedia(slot="video", name="clip.mp4", duration=120.5)
            gallery.media["image"] = image
            gallery.media["video"] = video

            session.add(gallery)
            session.flush()
            gallery.revalidate()
            image.revalidate()
            video.revalidate()

            assert gallery.id is not None
            assert image.id is not None
            assert video.id is not None
            assert len(gallery.media) == 2

    def test_create_gallery_empty(self, engine: Engine):
        """Test creating a Gallery with no media."""
        with Session(engine) as session:
            gallery = Gallery(name="Empty Gallery")
            session.add(gallery)
            session.flush()
            gallery.revalidate()

            assert gallery.id is not None
            assert len(gallery.media) == 0

    def test_add_media_to_existing_gallery(self, engine: Engine):
        """Test adding media to an existing Gallery."""
        with Session(engine) as session:
            gallery = Gallery(name="Growing Gallery")
            session.add(gallery)
            session.flush()

            image = ImageMedia(slot="image", name="added.png", width=800, height=600)
            gallery.media["image"] = image
            session.flush()
            image.revalidate()

            assert image.id is not None
            assert len(gallery.media) == 1
            assert gallery.media["image"].name == "added.png"

    def test_remove_media_from_gallery(self, engine: Engine):
        """Test removing media from a Gallery."""
        with Session(engine) as session:
            gallery = Gallery(name="Shrinking Gallery")
            image = ImageMedia(slot="image", name="temp.png", width=100, height=100)
            gallery.media["image"] = image
            session.add(gallery)
            session.flush()

            assert len(gallery.media) == 1
            del gallery.media["image"]
            session.flush()

            assert len(gallery.media) == 0

    def test_replace_media_in_gallery(self, engine: Engine):
        """Test replacing media at an existing key."""
        with Session(engine) as session:
            gallery = Gallery(name="Replace Gallery")
            original = ImageMedia(
                slot="image", name="original.png", width=640, height=480
            )
            gallery.media["image"] = original
            session.add(gallery)
            session.flush()

            replacement = ImageMedia(
                slot="image", name="replacement.png", width=1920, height=1080
            )
            gallery.media["image"] = replacement
            session.flush()

            assert gallery.media["image"].name == "replacement.png"

    def test_clear_all_media(self, engine: Engine):
        """Test clearing all media from a gallery."""
        with Session(engine) as session:
            gallery = Gallery(name="Clearable Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="img.png", width=100, height=100
            )
            gallery.media["video"] = VideoMedia(
                slot="video", name="vid.mp4", duration=30.0
            )
            session.add(gallery)
            session.flush()

            assert len(gallery.media) == 2
            gallery.media.clear()
            session.flush()

            assert len(gallery.media) == 0


class TestTypedRelationMapLoading:
    """Test lazy and eager loading of typed dict-based relationships."""

    def test_lazy_load_media(self, engine: Engine):
        """Test that media items are lazily loaded when accessed."""
        with Session(engine) as session:
            gallery = Gallery(name="Lazy Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="lazy_img.png", width=800, height=600
            )
            gallery.media["video"] = VideoMedia(
                slot="video", name="lazy_vid.mp4", duration=45.0
            )
            session.add(gallery)
            session.flush()
            gallery.revalidate()
            gallery_id = gallery.id
            session.commit()

        # New session to test lazy loading
        with Session(engine) as session:
            stmt = select(Gallery).where(Gallery["id"] == gallery_id)
            gallery = session.scalars(stmt).unique().one()
            assert isinstance(gallery, Gallery)
            assert isinstance(gallery.media, TypedRelationMap)

            # Accessing media triggers lazy load
            assert len(gallery.media) == 2
            assert gallery.media["image"].name == "lazy_img.png"
            assert gallery.media["video"].name == "lazy_vid.mp4"

    def test_eager_load_selectinload(self, engine: Engine):
        """Test eager loading media with selectinload."""
        with Session(engine) as session:
            gallery = Gallery(name="Eager Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="eager_img.png", width=400, height=300
            )
            gallery.media["video"] = VideoMedia(
                slot="video", name="eager_vid.mp4", duration=90.0
            )
            session.add(gallery)
            session.flush()
            gallery.revalidate()
            gallery_id = gallery.id
            session.commit()

        with Session(engine) as session:
            stmt = (
                select(Gallery)
                .where(Gallery["id"] == gallery_id)
                .options(selectinload(Gallery["media"]))
            )
            gallery = session.scalars(stmt).unique().one()
            assert isinstance(gallery, Gallery)
            assert len(gallery.media) == 2

    def test_load_preserves_keys(self, engine: Engine):
        """Test that dict keys are preserved after loading from DB."""
        with Session(engine) as session:
            gallery = Gallery(name="Key Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="key_img.png", width=100, height=100
            )
            gallery.media["video"] = VideoMedia(
                slot="video", name="key_vid.mp4", duration=10.0
            )
            session.add(gallery)
            session.flush()
            gallery.revalidate()
            gallery_id = gallery.id
            session.commit()

        with Session(engine) as session:
            stmt = select(Gallery).where(Gallery["id"] == gallery_id)
            gallery = session.scalars(stmt).unique().one()
            assert set(gallery.media.keys()) == {"image", "video"}

    def test_load_preserves_polymorphic_types(self, engine: Engine):
        """Test that polymorphic subtypes are preserved after loading from DB."""
        with Session(engine) as session:
            gallery = Gallery(name="Poly Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="poly_img.png", width=1024, height=768
            )
            gallery.media["video"] = VideoMedia(
                slot="video", name="poly_vid.mp4", duration=60.0
            )
            session.add(gallery)
            session.flush()
            gallery.revalidate()
            gallery_id = gallery.id
            session.commit()

        with Session(engine) as session:
            stmt = select(Gallery).where(Gallery["id"] == gallery_id)
            gallery = session.scalars(stmt).unique().one()
            assert isinstance(gallery.media["image"], ImageMedia)
            assert isinstance(gallery.media["video"], VideoMedia)
            assert gallery.media["image"].width == 1024
            assert gallery.media["video"].duration == 60.0


class TestTypedRelationMapSerialization:
    """Test serialization of TypedRelationMap after ORM operations."""

    def test_model_dump_after_create(self, engine: Engine):
        """Test model_dump after creating gallery with media."""
        with Session(engine) as session:
            gallery = Gallery(name="Dump Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="dump_img.png", width=640, height=480
            )
            gallery.media["video"] = VideoMedia(
                slot="video", name="dump_vid.mp4", duration=15.0
            )
            session.add(gallery)
            session.flush()

            data = gallery.model_dump()
            assert isinstance(data["media"], dict)
            assert set(data["media"].keys()) == {"image", "video"}
            assert data["media"]["image"]["name"] == "dump_img.png"
            assert data["media"]["image"]["width"] == 640
            assert data["media"]["video"]["duration"] == 15.0

    def test_model_dump_after_load(self, engine: Engine):
        """Test model_dump after loading from DB with selectinload."""
        with Session(engine) as session:
            gallery = Gallery(name="Load Dump Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="ld_img.png", width=320, height=240
            )
            session.add(gallery)
            session.flush()
            gallery.revalidate()
            gallery_id = gallery.id
            session.commit()

        with Session(engine) as session:
            stmt = (
                select(Gallery)
                .where(Gallery["id"] == gallery_id)
                .options(selectinload(Gallery["media"]))
            )
            gallery = session.scalars(stmt).unique().one()
            data = gallery.model_dump()
            assert isinstance(data["media"], dict)
            assert "image" in data["media"]
            assert data["media"]["image"]["name"] == "ld_img.png"

    def test_model_dump_empty_media(self, engine: Engine):
        """Test model_dump with empty media dict."""
        with Session(engine) as session:
            gallery = Gallery(name="Empty Dump Gallery")
            session.add(gallery)
            session.flush()

            data = gallery.model_dump()
            assert data["media"] == {}


class TestTypedRelationMapDictOperations:
    """Test dict operations through the ORM."""

    def test_get_existing_key(self, engine: Engine):
        """Test get() on a loaded gallery."""
        with Session(engine) as session:
            gallery = Gallery(name="Get Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="get_img.png", width=100, height=100
            )
            session.add(gallery)
            session.flush()

            assert gallery.media.get("image") is not None
            assert gallery.media["image"].name == "get_img.png"
            assert gallery.media.get("video") is None

    def test_pop_from_gallery(self, engine: Engine):
        """Test pop() removes and returns the media item."""
        with Session(engine) as session:
            gallery = Gallery(name="Pop Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="pop_img.png", width=100, height=100
            )
            session.add(gallery)
            session.flush()

            item = gallery.media.pop("image")
            assert item.name == "pop_img.png"
            assert len(gallery.media) == 0

    def test_keys_values_items(self, engine: Engine):
        """Test keys(), values(), items() views."""
        with Session(engine) as session:
            gallery = Gallery(name="Views Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="v_img.png", width=100, height=100
            )
            gallery.media["video"] = VideoMedia(
                slot="video", name="v_vid.mp4", duration=5.0
            )
            session.add(gallery)
            session.flush()

            assert set(gallery.media.keys()) == {"image", "video"}
            assert len(list(gallery.media.values())) == 2
            assert len(list(gallery.media.items())) == 2

    def test_contains_key(self, engine: Engine):
        """Test key membership check."""
        with Session(engine) as session:
            gallery = Gallery(name="Contains Gallery")
            gallery.media["image"] = ImageMedia(
                slot="image", name="c_img.png", width=100, height=100
            )
            session.add(gallery)
            session.flush()

            assert "image" in gallery.media
            assert "video" not in gallery.media

    def test_update_dict(self, engine: Engine):
        """Test update() with new media items."""
        with Session(engine) as session:
            gallery = Gallery(name="Update Gallery")
            session.add(gallery)
            session.flush()

            gallery.media.update(
                {
                    "image": ImageMedia(
                        slot="image", name="u_img.png", width=200, height=200
                    ),
                    "video": VideoMedia(slot="video", name="u_vid.mp4", duration=25.0),
                }
            )
            session.flush()

            assert len(gallery.media) == 2
            assert gallery.media["image"].name == "u_img.png"

    def test_setdefault(self, engine: Engine):
        """Test setdefault() inserts only if key is absent."""
        with Session(engine) as session:
            gallery = Gallery(name="Setdefault Gallery")
            existing = ImageMedia(
                slot="image", name="original.png", width=100, height=100
            )
            gallery.media["image"] = existing
            session.add(gallery)
            session.flush()

            new = ImageMedia(slot="image", name="new.png", width=200, height=200)
            result = gallery.media.setdefault("image", new)
            assert result.name == "original.png"


class TestTypedRelationMapBackReference:
    """Test back-reference from media item to Gallery."""

    def test_media_has_gallery_reference(self, engine: Engine):
        """Test that a media item can reference its parent Gallery."""
        with Session(engine) as session:
            gallery = Gallery(name="BackRef Gallery")
            image = ImageMedia(
                slot="image", name="backref_img.png", width=100, height=100
            )
            gallery.media["image"] = image
            session.add(gallery)
            session.flush()

            # Access via the ORM provider's back-reference
            assert image.__transmuter_provided__ is not None
            assert image.__transmuter_provided__.gallery.name == "BackRef Gallery"  # type: ignore
